#include "GS_Problem.h"
#include "ImageProcessing.h"
#include "ImageIO.h"

const extern char* ProbFilterTypeStr[PFT_LAST + 2] = 
{
	"PFT_SATURATE_COLOR", 
	"PFT_SHARPEN",	
	"PFT_DENOISE", 
	"PFT_LUM2RGB",
	"PFT_ROBUST_RESCALE",
	"PFT_EDGE_VIS1", 
	"PFT_EDGE_VIS2", 
	NULL
};

GS_Problem::GS_Problem(Params _params)
{
	this->params = _params;
	Resize(this->params.problemShape, this->params.plsParams.levelCount);
}

GS_Problem::GS_Problem(GS_ProblemInput& _input, Params _params)
{
	_params.problemShape = _input.Shape();
	this->params = _params;
	Resize(this->params.problemShape, this->params.plsParams.levelCount);
	ResetInput(_input);
}

void GS_Problem::Resize(GS_P_Shape shape, int levelCount)
{
	this->params.problemShape = shape;
	if(levelCount == -1)
		levelCount = this->params.plsParams.levelCount;
	else
		this->params.plsParams.levelCount = levelCount;

	this->input.ReAllocate(this->params.problemShape);
	CShape imgShape = this->params.problemShape.imgShape;
	this->result.ReAllocate(imgShape);

	//this->problemLevels.Resize(imgShape, this->params.plsParams.levelCount);
	this->problemLevels.Init(this->params.plsParams, imgShape);
}

void GS_Problem::ResetInput(const GS_ProblemInput& _input)
{
	INSIST(this->input.Shape() == _input.Shape());
	this->input.Copy(_input);
	verifyChannelsTypeCorrectness();
	PreProcessInput();
	this->result.channelsType = this->input.img.channelsType;
	ResetLevels();
}

void GS_Problem::ResetLevels()
{
	this->problemLevels.ResetLevels(this->input);
}

void GS_Problem::ResetResidual(const Indices &levelsNotToUse)
{
	this->problemLevels.ResetResidual(levelsNotToUse);
}

void GS_Problem::UpdateResult(const Indices &levelsToUse)
{
	this->result.ClearPixels();
	for(Indices::const_iterator iLevel = levelsToUse.begin(); 
		iLevel != levelsToUse.end(); 
		iLevel++)
	{
		GS_ProblemLevel* currLevel = this->problemLevels.GetLevel(*iLevel);
		ImageProcessing::Operate(this->result, this->result, currLevel->flt, ImageProcessing::OP_ADD);
	}

	make_resultUpdate_callbacks();
}


void GS_Problem::Copy(const GS_Problem& cpySrc, int srcChannel)
{
	//copy input
	this->input.Copy(cpySrc.input, srcChannel);	
	verifyChannelsTypeCorrectness();

	//copy result
	ImageProcessing::CopyChannel(this->result, cpySrc.result, 0, srcChannel);
	this->result.channelsType = ColorHelper::GetSubChannelType(cpySrc.result.channelsType, srcChannel);

	//copy levels
	this->problemLevels.Copy(cpySrc.problemLevels, srcChannel);
}

GS_Problem GS_Problem::ExtractChannel(int iChannel)
{
	CShape imgShape = this->params.problemShape.imgShape;
	CHECK_RANGE(iChannel, 0, imgShape.nBands - 1);

	Params newParams = this->params;
	newParams.problemShape.imgShape.nBands = 1;	
	GS_Problem newProb(newParams);
	newProb.Copy(*this, iChannel);

	return newProb;
}

GS_ProblemInput* GS_Problem::GetInputBuffer()
{
	return &this->input;
}

GS_Problem::Params GS_Problem::GetParams()
{
	return this->params;
}

GS_ProblemLevel* GS_Problem::GetLevel(int iLevel)
{
	return this->problemLevels.GetLevel(iLevel);
}

GS_Image GS_Problem::GetResult() 
{
	return this->result;
}

GS_Image GS_Problem::GetResultRGB(bool postProcessResult) const
{
	GS_Image resultRgb;

	if(this->input.img.Shape().nBands == 3)
		resultRgb = ColorHelper::ConvertChannels(CT_RGB, this->result);
	else
		resultRgb = ColorHelper::ConvertChannels(CT_LUM, this->result);

	FilterAuxData auxData;
	auxData.emData = this->input.emData;
	auxData.imgFN  = this->input.imgFN;

	if(postProcessResult)
		GS_Problem::FilterImg(resultRgb, this->params.postprocessFilters, auxData);

	return resultRgb;
}


void GS_Problem::verifyChannelsTypeCorrectness(bool tolerateUnknown)
{
	bool correct = false;

	ChannelsType inputCT = this->input.img.channelsType;
	if(tolerateUnknown == false)
	{
		INSIST(inputCT != CT_UNKNOWN);
	}
	if((inputCT == this->params.channelsType) ||
	   (inputCT == ColorHelper::GetSubChannelType(this->params.channelsType, 0, true)) ||
	   (inputCT == ColorHelper::GetSubChannelType(this->params.channelsType, 1, true)) ||
	   (inputCT == ColorHelper::GetSubChannelType(this->params.channelsType, 2, true)))
	{
		correct = true;
	}

	INSIST(correct == true);
}


void 
GS_Problem::Add_resultUpdate_callback(Callback callback)
{
	this->resultUpdate_Callbacks.push_back(callback);
}

void 
GS_Problem::make_resultUpdate_callbacks()
{
	for(uint iCallback = 0; iCallback < this->resultUpdate_Callbacks.size(); iCallback++)
	{
		Callback callback = this->resultUpdate_Callbacks[iCallback];
		callback.second(callback.first);
	}
}
