#include "GS_ProblemInput.h"
#include "ImageProcessing.h"
#include "ImageIO.h"
#include "FlowHelper.h"
#include "Disk.h"
#include "ColorHelper.h"

GS_ProblemInput::GS_ProblemInput(string _imgFN, Params loadParams, ChannelsType channelsType)
{
	this->imgFN = _imgFN;
	
	GS_Image imgTemp;
	ImageIO::ReadFile(imgTemp, this->imgFN);

	this->img = ColorHelper::ConvertChannels(channelsType, imgTemp);
	this->shape = GS_P_Shape(this->img.Shape());

	setupDefaultData();	

	if(loadParams.loadEmData == true)
	{
		string emDataFN = EdgeMetrics::GetFileName(imgFN);
		if((loadParams.loadOnlyIfExists == false) ||
		   (Disk::DoesFileExist(emDataFN) == true))
		{
			this->emData.Load(emDataFN, this->img.Shape());
			INSIST(IsEdgeMetricsDefined() == true);
		}
	}

	if(loadParams.loadFlowField == true)
	{
		string flowFN = FlowHelper::GetFlowFN(imgFN);
		if((loadParams.loadOnlyIfExists == false) ||
		   (Disk::DoesFileExist(flowFN) == true))
		{
			FlowHelper::ReadFlowField(flowFN, this->flowField);
			INSIST(IsFlowFieldDefined() == true);
		}
	}
}

GS_ProblemInput::GS_ProblemInput(GS_P_Shape _shape)
{
	ReAllocate(_shape);
}

GS_ProblemInput::GS_ProblemInput(const GS_Image& _img, ChannelsType channelsType)
{
	ReAllocate(_img.Shape());
	this->img = ColorHelper::ConvertChannels(channelsType, _img);
	setupDefaultData();
}

GS_P_Shape GS_ProblemInput::Shape()  const
{
	return this->shape;
}

void GS_ProblemInput::ReAllocate(GS_P_Shape _shape)
{
	resize(_shape);
}

void GS_ProblemInput::Copy(const GS_ProblemInput &cpySrc)
{
	INSIST(this->Shape() == cpySrc.Shape());
	ImageProcessing::CopyChannels(this->img,   cpySrc.img);
	ImageProcessing::CopyChannels(this->data,  cpySrc.data);
	ImageProcessing::CopyChannels(this->dataW, cpySrc.dataW);
	this->imgFN       = cpySrc.imgFN;
	this->emData      = cpySrc.emData.Clone();	
	this->auxDataObjs = cpySrc.auxDataObjs;
	this->flowField   = cpySrc.flowField.Clone();
}


void GS_ProblemInput::Copy(const GS_ProblemInput& cpySrc, int srcChannel)
{
	INSIST(this->shape.imgShape.SameIgnoringNBands(cpySrc.shape.imgShape));
	CHECK_RANGE(srcChannel, 0, cpySrc.shape.imgShape.nBands - 1);
	INSIST(this->shape.imgShape.nBands == 1);

	ImageProcessing::CopyChannel(this->img,   cpySrc.img,   0, srcChannel);
	ImageProcessing::CopyChannel(this->data,  cpySrc.data,  0, srcChannel);
	ImageProcessing::CopyChannel(this->dataW, cpySrc.dataW, 0, srcChannel);
	this->imgFN       = cpySrc.imgFN;
	this->emData      = cpySrc.emData.Clone();	
	this->auxDataObjs = cpySrc.auxDataObjs;
	this->flowField   = cpySrc.flowField.Clone();
}

GS_ProblemInput GS_ProblemInput::ExtractChannel(int iChannel)
{
	CHECK_RANGE(iChannel, 0, this->shape.imgShape.nBands - 1);

	CShape maskShape = this->shape.imgShape;
	maskShape.nBands = 1;
	GS_P_Shape newShape(maskShape);
	GS_ProblemInput newProb(newShape);
	newProb.Copy(*this, iChannel);

	return newProb;
}

void GS_ProblemInput::resize(GS_P_Shape _shape)
{
	this->shape = _shape;

	CShape imgShape = this->shape.imgShape;
	this->img.ReAllocate(imgShape);
	this->data.ReAllocate(imgShape);
	this->dataW.ReAllocate(imgShape);
}

void GS_ProblemInput::AddAuxData(string dataName, void *auxData, bool overwriteData)
{
	if((overwriteData == false) && IsAuxDataDefined(dataName))
	{
		REPORT_FAILURE("Trying to overwrite an already existing aux data - %s\n", dataName.c_str());
	}
	this->auxDataObjs[dataName] = auxData;
}

void* GS_ProblemInput::GetAuxData(string dataName)
{
	HashedPtrs::iterator findResult = this->auxDataObjs.find(dataName);
	INSIST(findResult != this->auxDataObjs.end());
	return findResult->second;
}

bool GS_ProblemInput::IsAuxDataDefined(string dataName)
{
	if(this->auxDataObjs.find(dataName) == this->auxDataObjs.end())
		return false;
	else
		return true;
}



bool GS_ProblemInput::IsEdgeMetricsDefined() const
{
	if(this->emData.edgeDirc.Shape().NodeCount() == 0)
		return false;
	else
	{
		CShape imgShape = this->img.Shape();
		CShape emShape(imgShape.width, imgShape.height, 1);

		INSIST(this->emData.edgeDirc.Shape()     == emShape);
		INSIST(this->emData.edgeLen.Shape()      == emShape);
		INSIST(this->emData.edgeStrength.Shape() == emShape);

		return true;
	}
}



bool GS_ProblemInput::IsFlowFieldDefined() const
{
	if(this->flowField.Shape().NodeCount() == 0)
		return false;
	else
	{
		CShape imgShape = this->img.Shape();
		CShape flowShape(imgShape.width, imgShape.height, 3);
		INSIST(this->flowField.Shape() == flowShape);
		return true;
	}
}


void GS_ProblemInput::setupDefaultData()
{
	this->data.ReAllocate(this->shape.imgShape);
	ImageProcessing::CopyChannels(this->data, this->img);

	this->dataW.ReAllocate(this->shape.imgShape);
	ImageProcessing::Operate(this->dataW, this->dataW, 1.0f, ImageProcessing::OP_SET);	
}

