#include "GS_ProblemLevels.h"
#include "ImageProcessing.h"
#include "ImageIO.h"

GS_ProblemLevels::GS_ProblemLevels(Params _params, CShape _imgShape)
{
	Init(_params, _imgShape);
}

void GS_ProblemLevels::Init(Params _params, CShape _imgShape)
{
	this->params = _params;
	INSIST(this->params.levelCount >= 1);
	Resize(_imgShape, this->params.levelCount);
}

void GS_ProblemLevels::Resize(CShape _imgShape, int levelCount)
{
	if(levelCount == -1)
	{
		INSIST(this->params.levelCount >= 1);
		levelCount = this->params.levelCount;
	}
	else
	{
		INSIST(levelCount >= 1);
		this->params.levelCount = levelCount;
	}
	this->levels.resize(this->params.levelCount);

	this->imgShape = _imgShape;	
	for(vector<GS_ProblemLevel>::iterator currLevel = this->levels.begin();
		currLevel != this->levels.end();
		currLevel++)
	{
		currLevel->Resize(this->imgShape);
	}

	if(levelCount >= GS_ProblemLevels::MinLevelsInProbForResidue)
		this->residualLevel.Resize(this->imgShape);
}

void GS_ProblemLevels::ResetLevels(GS_ProblemInput &input)
{
	INSIST(this->params.levelCount >= 1);

	GS_Image inputImg = input.img;

	GS_Image fineImg = this->levels[0].org;
	ImageProcessing::CopyChannels(fineImg, inputImg);

	for(int iLevel = 1; iLevel < this->params.levelCount; iLevel++)
	{
		GS_Image finerImg   = this->levels[iLevel - 1].org;
		GS_Image coarserImg = this->levels[iLevel].org;	

		float levelSigmaD = this->params.sigmaD * pow(this->params.sigmaD_mult, iLevel - 1.0f);
		float levelSigmaS = this->params.sigmaS * pow(this->params.sigmaS_mult, iLevel - 1.0f);
		int kernelSize = (int) max(5.0f, levelSigmaD * 2);
		kernelSize += (kernelSize % 2 == 0) ? 1 : 0; 
		GS_Image filteredImg;
		filteredImg.channelsType = inputImg.channelsType;

		switch(this->params.levelsType)
		{
		case LT_DOG:
			finerImg.channelsType = inputImg.channelsType;
			filteredImg = ImageProcessing::GaussianBlur(finerImg, 
														kernelSize,
													    levelSigmaD);
			break;
		case LT_DOB:
			filteredImg = ImageProcessing::BilateralFilter(finerImg, 
														   levelSigmaD,
														   levelSigmaS);
			break;
		default:
			REPORT_FAILURE("Unknown levels type - %i\n", this->params.levelsType);
			break;
		}

		ImageProcessing::CopyChannels(coarserImg, filteredImg);
		ImageProcessing::Operate(finerImg, finerImg, coarserImg, ImageProcessing::OP_SUB);		
	}


	for(vector<GS_ProblemLevel>::iterator currLevel = this->levels.begin();
		currLevel != this->levels.end();
		currLevel++)
	{
		currLevel->ComputeGrads();
		ImageProcessing::CopyChannels(currLevel->flt,   currLevel->org);
		ImageProcessing::CopyChannels(currLevel->data,  currLevel->org);
		ImageProcessing::CopyChannels(currLevel->dataW, input.dataW);
		ImageProcessing::Operate(currLevel->dataW, currLevel->dataW, 1.0f / this->params.levelCount, ImageProcessing::OP_DIV);
	}

	for(int iLevel = -1; iLevel < this->params.levelCount; iLevel++)
	{
		if((iLevel >= 0) ||
		   (GetLevelCount() >= GS_ProblemLevels::MinLevelsInProbForResidue))
		{
			GS_ProblemLevel* currLevel = GetLevel(iLevel);
			currLevel->org.channelsType     = input.img.channelsType;
			currLevel->flt.channelsType     = input.img.channelsType;
			currLevel->data.channelsType    = input.data.channelsType;
			currLevel->dataW.channelsType   = input.dataW.channelsType;
			currLevel->orgGF_X.channelsType = input.img.channelsType;
			currLevel->orgGF_Y.channelsType = input.img.channelsType;
			currLevel->fltGF_X.channelsType = input.img.channelsType;
			currLevel->fltGF_Y.channelsType = input.img.channelsType;
		}
	}
}

void GS_ProblemLevels::ResetResidual(const Indices &levelsNotToUse)
{
	INSIST(GetLevelCount() >= GS_ProblemLevels::MinLevelsInProbForResidue);
	INSIST(levelsNotToUse.find(-1) == levelsNotToUse.end());
	CHECK_RANGE((int) levelsNotToUse.size(), 1, GetLevelCount() - GS_ProblemLevels::MinLevelsInResidue);
	
	this->residualLevel.org.ClearPixels();
	this->residualLevel.data.ClearPixels();

	for(int iLevel = 0; iLevel < GetLevelCount(); iLevel++)
	{
		if(levelsNotToUse.find(iLevel) == levelsNotToUse.end())
		{
			GS_ProblemLevel *problemLevel = this->GetLevel(iLevel);
			ImageProcessing::Operate(this->residualLevel.org,   this->residualLevel.org,   problemLevel->org,   ImageProcessing::OP_ADD);
			ImageProcessing::Operate(this->residualLevel.data,  this->residualLevel.data,  problemLevel->data,  ImageProcessing::OP_ADD);
			ImageProcessing::Operate(this->residualLevel.dataW, this->residualLevel.dataW, problemLevel->dataW, ImageProcessing::OP_ADD);
		}
	}

	this->residualLevel.ComputeGrads();
	ImageProcessing::CopyChannels(this->residualLevel.flt, this->residualLevel.org);
}


GS_ProblemLevel* GS_ProblemLevels::GetLevel(int iLevel) 
{
	CHECK_RANGE(iLevel, -1, GetLevelCount() - 1);

	if(iLevel == -1)
	{
		INSIST(GetLevelCount() >= GS_ProblemLevels::MinLevelsInProbForResidue);
		return &this->residualLevel;
	}
	else
	{
		return &this->levels[iLevel];
	}
}


void GS_ProblemLevels::Copy(const GS_ProblemLevels& cpySrc, int srcChannel)
{
	INSIST(this->GetLevelCount() == cpySrc.GetLevelCount());
	INSIST(this->imgShape.SameIgnoringNBands(cpySrc.imgShape));
	CHECK_RANGE(srcChannel, 0, cpySrc.imgShape.nBands - 1);
	INSIST(this->imgShape.nBands == 1);
	INSIST(this->params == cpySrc.params);

	if(GetLevelCount() >= GS_ProblemLevels::MinLevelsInProbForResidue)
		this->residualLevel.Copy(cpySrc.residualLevel, srcChannel);

	for(int iLevel = 0; iLevel < this->GetLevelCount(); iLevel++)
	{
		this->levels[iLevel].Copy(cpySrc.levels[iLevel], srcChannel);
	}
}

void GS_ProblemLevels::Write(string prefix, string fileExt, Indices levelsToUse) 
{
	INSIST(this->GetLevelCount() > 0);
	if(levelsToUse.size() == 0)
	{
		for(int iLevel = 0; iLevel < this->GetLevelCount(); iLevel++)
		{
			levelsToUse.insert(iLevel);
		}
	}

	int levelFileID = 0;
	for(Indices::const_iterator currLevel = levelsToUse.begin();
		currLevel != levelsToUse.end();
		currLevel++, levelFileID++)
	{
		this->GetLevel(*currLevel)->Write(prefix, levelFileID, fileExt); 
	}
}


void GS_ProblemLevels::Read(string prefix, string fileExt)
{
	INSIST(this->GetLevelCount() > 0);
	for(int iLevel = 0; iLevel < this->GetLevelCount(); iLevel++)
	{
		this->GetLevel(iLevel)->Read(prefix, iLevel, fileExt);
	}
}

