#include "GS_FourierSolver.h"
#include "Disk.h"
#include "ColorHelper.h"
#include "ImageIO.h"
#include "Timer.h"
#define NUM_OF_THREADS 1


GS_FourierSolver::GS_FourierSolver(GS_Problem* _problem, Params _params)
{
	this->params  = _params;
	this->problem = _problem;

	GS_Problem::Params problemParams = this->problem->GetParams();
	this->problemShape = problemParams.problemShape;
	CShape imgShape = this->problemShape.imgShape;
	INSIST(imgShape.PixelCount() > 0);
	INSIST(this->problem->GetLevelCount() > 0);	
	INSIST(problemParams.plsParams.levelCount == this->problem->GetLevelCount());	

	adjustParams();

	initFFTW();
}


void 
GS_FourierSolver::adjustParams()
{
	GS_Problem::Params problemParams = this->problem->GetParams();
	CShape imgShape = this->problemShape.imgShape;
	
	this->params.pinchFuncScale = ColorHelper::GetGradScale(problemParams.channelsType);
	
	if(this->params.levelsToSolve.size() == 0)
	{
		for(int iLevel = 0; iLevel < this->problem->GetLevelCount(); iLevel++)
		{
			this->params.levelsToSolve.insert(iLevel);
		}
	}
	else
	{
		int levelsSelected = (int) this->params.levelsToSolve.size();

		for(Indices::const_iterator iLevel = this->params.levelsToSolve.begin();
			iLevel != this->params.levelsToSolve.end();
			iLevel++)
		{
			CHECK_RANGE(*iLevel, 0, this->problem->GetLevelCount() - 1);
		}

		if((this->problem->GetLevelCount() - levelsSelected) >= GS_ProblemLevels::MinLevelsInResidue)
		{
			this->problem->ResetResidual(this->params.levelsToSolve);
			this->params.levelsToSolve.insert(-1);
		}
		else
		{
			for(int iLevel = 0; iLevel < this->problem->GetLevelCount(); iLevel++)
			{
				this->params.levelsToSolve.insert(iLevel);
			}
		}
	}
	
	if(this->params.channelsToSolve.size() == 0)
	{		
		for(int iChannel = 0; iChannel < imgShape.nBands; iChannel++)
		{
			this->params.channelsToSolve.insert(iChannel);
		}
	}	
	else
	{
		for(Indices::const_iterator iChannel = this->params.channelsToSolve.begin();
			iChannel != this->params.channelsToSolve.end();
			iChannel++)
		{
			CHECK_RANGE(*iChannel, 0, imgShape.nBands - 1);
		}
	}
}


GS_FourierSolver::~GS_FourierSolver()
{
	FILE* outFile = fopen("fftw_wisdom.data", "w");
	INSIST(outFile != NULL);
	fftwf_export_wisdom_to_file(outFile);
	fclose(outFile);

	destroyFFTW();
}

void
GS_FourierSolver::initFFTW()
{
	FILE* inFile = fopen("fftw_wisdom.data", "r");
	if(inFile != NULL)
	{
		fftwf_import_wisdom_from_file(inFile);
		fclose(inFile);
	}

	CShape imgShape = this->problemShape.imgShape;
	int nodeCount = imgShape.NodeCount();
	INSIST(nodeCount > 0);
	
	this->fftBuff   = (float*) fftwf_malloc(sizeof(*fftBuff) * nodeCount);
	INSIST(this->fftBuff != NULL);

	this->ftLapY = (float*) fftwf_malloc(sizeof(*ftLapY) * imgShape.height);
	this->ftLapX = (float*) fftwf_malloc(sizeof(*ftLapX) * imgShape.width);
	INSIST(ftLapX != NULL);
	INSIST(ftLapY != NULL);
	for(int x = 0; x < imgShape.width; x++)
	{
		this->ftLapX[x] = 2.0f * cos(M_PI_FLOAT * x / (imgShape.width - 1));
	}
	for(int y = 0; y < imgShape.height; y++)
	{
		this->ftLapY[y] = -4.0f + (2.0f * cos(M_PI_FLOAT * y / (imgShape.height - 1)));
	}

	//bug
	//fftwf_plan_with_nthreads(NUM_OF_THREADS);
	
	Timer timer;
	printf("FFTW planning...\n");
	this->fftPlan = fftwf_plan_r2r_2d(imgShape.height, imgShape.width, 
								      fftBuff, fftBuff, 
								      FFTW_REDFT00, FFTW_REDFT00, FFTW_MEASURE);
	timer.DumpSeconds("\tDone - ");
}

void
GS_FourierSolver::destroyFFTW()
{

	fftwf_free(this->fftBuff);
	fftwf_free(this->ftLapX);
	fftwf_free(this->ftLapY);
	fftwf_destroy_plan(this->fftPlan);
}

void GS_FourierSolver::Solve()
{
	INSIST(this->params.channelsToSolve.size() > 0);
	INSIST(this->params.levelsToSolve.size()   > 0);


	for(Indices::const_iterator iLevel = this->params.levelsToSolve.begin();
			iLevel != this->params.levelsToSolve.end();
			iLevel++)
	{
		solve(*iLevel);
	}

	this->problem->UpdateResult(this->params.levelsToSolve);

	if(this->params.writeProb || this->params.writeSol)
	{
		string fileName = this->problem->GetInputBuffer()->GetFileName();
		string fileID   = this->params.outDir + Disk::GetFileNameID(Disk::GetBaseFileName(fileName));

		if(this->params.writeProb)
		{
			for(Indices::const_iterator iChannel = this->params.channelsToSolve.begin();
				iChannel != this->params.channelsToSolve.end();
				iChannel++)
			{
				string channelID = this->params.outDir + 
								   CImage::GetChannelStr(*iChannel) + "-" + 
								   Disk::GetFileNameID(Disk::GetBaseFileName(fileName));
				this->problem->ExtractChannel(*iChannel).Write(channelID, "raw", this->params.levelsToSolve);	
			}			
		}
		
		//this->problem->Write(fileID, "tga");
		this->problem->WriteSol(fileID, "tga");
	}
}


void GS_FourierSolver::solve(int iLevel)
{
	printf("Solving level - %i\n", iLevel); 

	CShape imgShape = this->problemShape.imgShape;
	int nodeCount = imgShape.NodeCount();
	INSIST(nodeCount > 0);

	GS_ProblemLevel* levelProb = this->problem->GetLevel(iLevel);
	GS_Image levelOrg,  levelFlt;
	GS_Image levelData, levelDataW;
	GS_Image levelOrgGF_X, levelOrgGF_Y;
	GS_Image levelFltGF_X, levelFltGF_Y;
	levelProb->GetData(levelOrg, levelFlt, 
					   levelData, levelDataW,
					   levelOrgGF_X, levelOrgGF_Y, levelFltGF_X, levelFltGF_Y);


	for(Indices::const_iterator iChannel = this->params.channelsToSolve.begin();
		iChannel != this->params.channelsToSolve.end();
		iChannel++)
	{
		ImageProcessing::CopyChannel(levelFlt, levelData, *iChannel, *iChannel);
	}


	
	fftwSolve(levelFlt, levelFltGF_X, levelFltGF_Y, (float) this->params.fidelity_data);
}



//void 
//GS_FourierSolver::fftwSolve(CFloatImage& imgData, CFloatImage& imgGradX, CFloatImage& imgGradY, float dataCost)
//{
//	//int retVal = fftwf_init_threads();
//	//INSIST(retVal != 0);
//	Timer timer;
//
//	CShape imgShape = imgData.Shape();
//	int nodeCount   = imgShape.NodeCount();
//	INSIST(nodeCount > 0);
//
//	for(Indices::const_iterator iChannelRef = this->params.channelsToSolve.begin();
//		iChannelRef != this->params.channelsToSolve.end();
//		iChannelRef++)
//	{
//		int iChannel = *iChannelRef;
//
//		int nodeAddr        = 0;
//		int pixelAddr       = iChannel;
//		int rightPixelAddr  = imgShape.nBands + iChannel;
//		int topPixelAddr    = (imgShape.width * imgShape.nBands) + iChannel;
//		int leftPixelAddr   = -imgShape.nBands + iChannel;
//		int bottomPixelAddr = -(imgShape.width * imgShape.nBands) + iChannel;
//
//		float dcSum = 0.0f;
//
//		for(int y = 0; y < imgShape.height; y++)
//		for(int x = 0; x < imgShape.width;  x++, 
//			nodeAddr++, pixelAddr += imgShape.nBands, rightPixelAddr += imgShape.nBands, topPixelAddr += imgShape.nBands,
//			leftPixelAddr += imgShape.nBands, bottomPixelAddr += imgShape.nBands)
//		{
//			float dcMult = 1.0f;
//			if((x > 0) && (x < imgShape.width  - 1))
//				dcMult *= 2.0f;
//			if((y > 0) && (y < imgShape.height - 1))
//				dcMult *= 2.0f;
//			dcSum += dcMult * imgData[pixelAddr];
//
//			fftBuff[nodeAddr] = dataCost * imgData[pixelAddr];			
//
//			if((x > 0) && (x < imgShape.width - 1))
//				fftBuff[nodeAddr] -= (imgGradX[rightPixelAddr] - imgGradX[pixelAddr]);
//			else
//				fftBuff[nodeAddr] -= (-2.0f * imgGradX[pixelAddr]);
//
//			if((y > 0) && (y < imgShape.height - 1))
//				fftBuff[nodeAddr] -= (imgGradY[topPixelAddr] - imgGradY[pixelAddr]);
//			else
//				fftBuff[nodeAddr] -= (-2.0f * imgGradY[pixelAddr]);
//		}
//
//		fftwf_execute(fftPlan);
//		nodeAddr = 0;
//		for(int y = 0; y < imgShape.height; y++)
//		for(int x = 0; x < imgShape.width;  x++, nodeAddr++)
//		{
//			float ftLapResponse = ftLapY[y] + ftLapX[x]; 
//			fftBuff[nodeAddr] /= (dataCost - ftLapResponse);
//		}
//		fftBuff[0] = dcSum;
//		fftwf_execute(fftPlan);
//		
//		float fftDenom = 4.0f * (imgShape.width - 1) * (imgShape.height - 1);
//		pixelAddr = iChannel;
//		for(int iNode = 0; iNode < nodeCount; iNode++, pixelAddr += imgShape.nBands)
//		{
//			imgData[pixelAddr] = fftBuff[iNode] / fftDenom;	
//		}
//	}
//
//	timer.DumpSeconds("\tDone - ");
//}



void 
GS_FourierSolver::fftwSolve(CFloatImage& imgData, CFloatImage& imgGradX, CFloatImage& imgGradY, float dataCost)
{
	//int retVal = fftwf_init_threads();
	//INSIST(retVal != 0);
	Timer timer;

	CShape imgShape = imgData.Shape();
	int nodeCount   = imgShape.NodeCount();
	INSIST(nodeCount > 0);

	CFloatImage vis(imgShape.width * 2, imgShape.height * 2, imgShape.nBands);
	CFloatImage vis2(imgShape.width * 2, imgShape.height * 2, imgShape.nBands);
	CFloatImage vis3(imgShape);
	CFloatImage vis4(imgShape.width * 2, imgShape.height * 2, imgShape.nBands);
	vis.ClearPixels();
	vis2.ClearPixels();
	vis3.ClearPixels();
	vis4.ClearPixels();

	for(Indices::const_iterator iChannelRef = this->params.channelsToSolve.begin();
		iChannelRef != this->params.channelsToSolve.end();
		iChannelRef++)
	{
		int iChannel = *iChannelRef;

		int nodeAddr        = 0;
		int pixelAddr       = iChannel;
		int rightPixelAddr  = imgShape.nBands + iChannel;
		int topPixelAddr    = (imgShape.width * imgShape.nBands) + iChannel;
		int leftPixelAddr   = -imgShape.nBands + iChannel;
		int bottomPixelAddr = -(imgShape.width * imgShape.nBands) + iChannel;

		float dcSum = 0.0f;

		for(int y = 0; y < imgShape.height; y++)
		for(int x = 0; x < imgShape.width;  x++, 
			nodeAddr++, pixelAddr += imgShape.nBands, rightPixelAddr += imgShape.nBands, topPixelAddr += imgShape.nBands,
			leftPixelAddr += imgShape.nBands, bottomPixelAddr += imgShape.nBands)
		{
			float dcMult = 1.0f;
			if((x > 0) && (x < imgShape.width  - 1))
				dcMult *= 2.0f;
			if((y > 0) && (y < imgShape.height - 1))
				dcMult *= 2.0f;
			dcSum += dcMult * imgData[pixelAddr];

			fftBuff[nodeAddr] = dataCost * imgData[pixelAddr];			

			if((x > 0) && (x < imgShape.width - 1))
				fftBuff[nodeAddr] -= (imgGradX[rightPixelAddr] - imgGradX[pixelAddr]);
			else
				fftBuff[nodeAddr] -= (-2.0f * imgGradX[pixelAddr]);

			if((y > 0) && (y < imgShape.height - 1))
				fftBuff[nodeAddr] -= (imgGradY[topPixelAddr] - imgGradY[pixelAddr]);
			else
				fftBuff[nodeAddr] -= (-2.0f * imgGradY[pixelAddr]);


			vis3.Pixel(x, y, iChannel) = fftBuff[nodeAddr] + 0.5f;
		}

		fftwf_execute(fftPlan);
		nodeAddr = 0;
		for(int y = 0; y < imgShape.height; y++)
		for(int x = 0; x < imgShape.width;  x++, nodeAddr++)
		{
			vis.Pixel(imgShape.width - x, imgShape.height - y, iChannel) = fftBuff[nodeAddr];
			vis.Pixel(imgShape.width + x, imgShape.height + y, iChannel) = fftBuff[nodeAddr];
			vis.Pixel(imgShape.width + x, imgShape.height - y, iChannel) = fftBuff[nodeAddr];
			vis.Pixel(imgShape.width - x, imgShape.height + y, iChannel) = fftBuff[nodeAddr];


			float ftLapResponse = ftLapY[y] + ftLapX[x]; 
			fftBuff[nodeAddr] /= (dataCost - ftLapResponse);

			vis2.Pixel(imgShape.width - x, imgShape.height - y, iChannel) =  1.0f / (dataCost - ftLapResponse);
			vis2.Pixel(imgShape.width + x, imgShape.height + y, iChannel) =  1.0f / (dataCost - ftLapResponse);
			vis2.Pixel(imgShape.width + x, imgShape.height - y, iChannel) =  1.0f / (dataCost - ftLapResponse);
			vis2.Pixel(imgShape.width - x, imgShape.height + y, iChannel) =  1.0f / (dataCost - ftLapResponse);

			fftBuff[nodeAddr] = 1.0f / (dataCost - ftLapResponse);
		}
		//fftBuff[0] = dcSum;
		//fftBuff[0] = dcSum;
		fftwf_execute(fftPlan);

		nodeAddr = 0;
		for(int y = 0; y < imgShape.height; y++)
		for(int x = 0; x < imgShape.width;  x++, nodeAddr++)
		{
			vis4.Pixel(imgShape.width - x, imgShape.height - y, iChannel) = fftBuff[nodeAddr] / 1.0f;
			vis4.Pixel(imgShape.width + x, imgShape.height + y, iChannel) = fftBuff[nodeAddr] / 1.0f;
			vis4.Pixel(imgShape.width + x, imgShape.height - y, iChannel) = fftBuff[nodeAddr] / 1.0f;
			vis4.Pixel(imgShape.width - x, imgShape.height + y, iChannel) = fftBuff[nodeAddr] / 1.0f;
		}
		
		float fftDenom = 4.0f * (imgShape.width - 1) * (imgShape.height - 1);
		pixelAddr = iChannel;
		for(int iNode = 0; iNode < nodeCount; iNode++, pixelAddr += imgShape.nBands)
		{
			imgData[pixelAddr] = fftBuff[iNode] / fftDenom;	
		}
	}

	ImageIO::WriteFile(vis, "1.tga");
	ImageIO::WriteFile(vis2, "2.tga");
	ImageIO::WriteFile(vis3, "3.tga");
	ImageIO::WriteFile(vis4, "4.tga");
	//exit(0);

	timer.DumpSeconds("\tDone - ");
}


