#include "GS_Solver.h"
#include "Disk.h"
#include "ColorHelper.h"
#include "ImageIO.h"
#include "LinearSolverDefs.h"

GS_Solver::GS_Solver(GS_Problem* _problem, Params _params)
{
	this->params  = _params;
	this->problem = _problem;

	GS_Problem::Params problemParams = this->problem->GetParams();
	this->problemShape = problemParams.problemShape;
	CShape imgShape = this->problemShape.imgShape;
	INSIST(imgShape.PixelCount() > 0);
	INSIST(this->problem->GetLevelCount() > 0);	
	INSIST(problemParams.plsParams.levelCount == this->problem->GetLevelCount());	

	this->params.pinchFuncScale = ColorHelper::GetGradScale(problemParams.channelsType);
	
	if(this->params.levelsToSolve.size() == 0)
	{
		for(int iLevel = 0; iLevel < this->problem->GetLevelCount(); iLevel++)
		{
			this->params.levelsToSolve.insert(iLevel);
		}
	}
	else
	{
		int levelsSelected = (int) this->params.levelsToSolve.size();

		for(Indices::const_iterator iLevel = this->params.levelsToSolve.begin();
			iLevel != this->params.levelsToSolve.end();
			iLevel++)
		{
			CHECK_RANGE(*iLevel, 0, this->problem->GetLevelCount() - 1);
		}

		if((this->problem->GetLevelCount() - levelsSelected) >= GS_ProblemLevels::MinLevelsInResidue)
		{
			this->problem->ResetResidual(this->params.levelsToSolve);
			this->params.levelsToSolve.insert(-1);
		}
		else
		{
			for(int iLevel = 0; iLevel < this->problem->GetLevelCount(); iLevel++)
			{
				this->params.levelsToSolve.insert(iLevel);
			}
		}
	}
	
	if(this->params.channelsToSolve.size() == 0)
	{		
		for(int iChannel = 0; iChannel < imgShape.nBands; iChannel++)
		{
			this->params.channelsToSolve.insert(iChannel);
		}
	}	
	else
	{
		for(Indices::const_iterator iChannel = this->params.channelsToSolve.begin();
			iChannel != this->params.channelsToSolve.end();
			iChannel++)
		{
			CHECK_RANGE(*iChannel, 0, imgShape.nBands - 1);
		}
	}
}


void GS_Solver::Solve()
{
	INSIST(this->params.channelsToSolve.size() > 0);
	INSIST(this->params.levelsToSolve.size()   > 0);

	for(Indices::const_iterator iChannel = this->params.channelsToSolve.begin();
			iChannel != this->params.channelsToSolve.end();
			iChannel++)
	{
		solveChannel(*iChannel);
	}
	this->problem->UpdateResult(this->params.levelsToSolve);

	if(this->params.writeProb || this->params.writeSol)
	{
		string fileName = this->problem->GetInputBuffer()->GetFileName();
		string fileID   = this->params.outDir + Disk::GetFileNameID(Disk::GetBaseFileName(fileName));

		if(this->params.writeProb)
		{
			for(Indices::const_iterator iChannel = this->params.channelsToSolve.begin();
				iChannel != this->params.channelsToSolve.end();
				iChannel++)
			{
				string channelID = this->params.outDir + 
								   CImage::GetChannelStr(*iChannel) + "-" + 
								   Disk::GetFileNameID(Disk::GetBaseFileName(fileName));
				this->problem->ExtractChannel(*iChannel).Write(channelID, "raw", this->params.levelsToSolve);	
			}			
		}
		
		//this->problem->Write(fileID, "tga");
		this->problem->WriteSol(fileID, "tga");
	}	
}


void GS_Solver::solveChannel(int iChannel)
{
	CShape imgShape = this->problemShape.imgShape;

	uint numOfVars = imgShape.NodeCount() * (int) this->params.levelsToSolve.size();
	INSIST(numOfVars > 0);
	SparseSolverParams solverParams = SparseSolverParams::Default();
	solverParams.nonZeroInMatGuess = numOfVars * (4 + this->params.levelsToSolve.size());
	solverParams.usePrecondHB = true;
	solverParams.maxLevelsInPrecond = this->params.maxLevelsInPrecond;
	for(uint iLevel = 0; iLevel < this->params.levelsToSolve.size(); iLevel++)
	{
		solverParams.imgShapes.push_back(imgShape);
	}
	SparseSolver solver(numOfVars, solverParams);

	int iSolLevel = 0;
	for(Indices::const_iterator iProbLevel = this->params.levelsToSolve.begin();
		iProbLevel != this->params.levelsToSolve.end();
		iProbLevel++, iSolLevel++)
	{	
		defineChannelEqs(iChannel, *iProbLevel, iSolLevel, solver);		
	}

	solver.Solve(this->params.accuracy, this->params.maxIters, true);

	iSolLevel = 0;
	for(Indices::const_iterator iProbLevel = this->params.levelsToSolve.begin();
		iProbLevel != this->params.levelsToSolve.end();
		iProbLevel++, iSolLevel++)
	{
		GS_ProblemLevel* problemLevel = this->problem->GetLevel(*iProbLevel);
		GS_Image         levelFlt     = problemLevel->flt;

		uint nodeAddr  = iSolLevel * imgShape.NodeCount();
		uint pixelAddr = iChannel;
		for(int y = 0; y < imgShape.height; y++)
		for(int x = 0; x < imgShape.width;  x++, nodeAddr++, pixelAddr += imgShape.nBands)
		{
			double sol = solver.GetSolution(nodeAddr);
			levelFlt[pixelAddr] = (imgType) sol;
		}
	}
}

 /*
void GS_Solver::defineChannelEqs(int iChannel, int iProbLevel, int iSolLevel, SparseSolver &solver)
{
	CShape imgShape = this->problemShape.imgShape;
	CHECK_RANGE(iChannel, 0, imgShape.nBands - 1);

	const int EquationVars     = 5;
	const int dx[EquationVars] = { 0, -1,  0,  1,  0};
	const int dy[EquationVars] = {-1,  0,  0,  0,  1};

	int levelsToSolve = (int) this->params.levelsToSolve.size();
	int*    eqVarAddr = new int[EquationVars    + levelsToSolve];
	double* eqVarCoef = new double[EquationVars + levelsToSolve];

	GS_ProblemLevel* levelProb = this->problem->GetLevel(iProbLevel);
	GS_Image levelOrg,  levelFlt;
	GS_Image levelData, levelDataW;
	GS_Image levelOrgGF_X, levelOrgGF_Y;
	GS_Image levelFltGF_X, levelFltGF_Y;
	levelProb->GetData(levelOrg, levelFlt, 
					   levelData, levelDataW,
					   levelOrgGF_X, levelOrgGF_Y, levelFltGF_X, levelFltGF_Y);

	GS_Image dataImg, dataImgW;
	if(this->params.useGlobalDatacost)
	{
		dataImg  = this->problem->GetInputBuffer()->data;
		dataImgW = this->problem->GetInputBuffer()->dataW;
	}
	else
	{
		dataImg    = levelData;
		levelDataW = levelDataW;
	}

	uint nodeAddr  = iSolLevel * imgShape.NodeCount();
	uint pixelAddr = iChannel;
	for(int y = 0; y < imgShape.height; y++)
	for(int x = 0; x < imgShape.width;  x++, nodeAddr++, pixelAddr += imgShape.nBands)
	{
		double rhs = 0;
		double fxy = levelFlt[pixelAddr];

		int    eqVarCount   = 0;
		int    nodeVarIndex = -1;
		double nodeVarCoef  = 0.0;
		for(int iEqVar = 0; iEqVar < EquationVars; iEqVar++)
		{
			if(iEqVar == 2)
			{
				INSIST(dx[iEqVar] == 0);
				INSIST(dy[iEqVar] == 0);

				eqVarAddr[eqVarCount] = nodeAddr;
				nodeVarIndex = eqVarCount;
				eqVarCount++;

				nodeVarCoef += this->params.fidelity_data * dataImgW[pixelAddr];
				rhs         += this->params.fidelity_data * dataImgW[pixelAddr] * dataImg[pixelAddr];				

				continue;
			}

			int neighbourX = x + dx[iEqVar];
			int neighbourY = y + dy[iEqVar];
			if(imgShape.InBounds(neighbourX, neighbourY) == false) continue;
	
			uint neighbourNodeAddr  = imgShape.NodeIndex(neighbourX, neighbourY);
			uint neighbourPixelAddr = (neighbourNodeAddr * imgShape.nBands) + iChannel;
			neighbourNodeAddr      += iSolLevel * imgShape.NodeCount();
			double neighPinch;
			double neighGrad;
			double neighCoef;

			if(neighbourY != y)
			{
				INSIST(neighbourX == x);
				if(y > neighbourY)
				{
					neighGrad  = (double) levelFltGF_Y[pixelAddr];
					neighPinch = (double) myAbs(levelFltGF_Y[pixelAddr] - levelOrgGF_Y[pixelAddr]);
				}
				else
				{
					neighGrad  = (double) -levelFltGF_Y[neighbourPixelAddr];
					neighPinch = (double) myAbs(levelFltGF_Y[neighbourPixelAddr] - levelOrgGF_Y[neighbourPixelAddr]);
				}
			}
			else
			{
				INSIST(neighbourX != x);
				if(x > neighbourX)
				{
					neighGrad  = (double) levelFltGF_X[pixelAddr];
					neighPinch = (double) myAbs(levelFltGF_X[pixelAddr] - levelOrgGF_X[pixelAddr]);
				}
				else
				{
					neighGrad  = (double) -levelFltGF_X[neighbourPixelAddr];
					neighPinch = (double) myAbs(levelFltGF_X[neighbourPixelAddr] - levelOrgGF_X[neighbourPixelAddr]);
				}		
			}

			neighCoef = this->params.fidelity_sGrad / pow((this->params.pinchFuncScale * neighPinch) + 1.0, this->params.pinchFuncExp);
			if(neighCoef != 0.0f)
			{
				if(neighbourNodeAddr >= nodeAddr)
				{
					eqVarAddr[eqVarCount] = neighbourNodeAddr;
					eqVarCoef[eqVarCount] = -neighCoef;
					eqVarCount++;
				}

				rhs += neighCoef * neighGrad;
				nodeVarCoef += neighCoef;
			}
		}

		if(this->params.useGlobalDatacost)
		{
			for(int vNeigh = iSolLevel + 1; vNeigh < levelsToSolve; vNeigh++)
			{
				eqVarAddr[eqVarCount] = nodeAddr + ((vNeigh - iSolLevel) * imgShape.NodeCount());
				eqVarCoef[eqVarCount] = this->params.fidelity_data * dataImgW[pixelAddr];
				eqVarCount++;
			}
		}

		eqVarCoef[nodeVarIndex] = nodeVarCoef;
		ENSURE(nodeVarCoef > 0.0);
		solver.SetRow(nodeAddr, eqVarCount, fxy, rhs, eqVarAddr, eqVarCoef);
	}

	delete [] eqVarAddr;
	delete [] eqVarCoef;
}
//*/

void GS_Solver::defineChannelEqs(int iChannel, int iProbLevel, int iSolLevel, SparseSolver &solver)
{
	CShape imgShape = this->problemShape.imgShape;
	CHECK_RANGE(iChannel, 0, imgShape.nBands - 1);
	
	EdgeMetrics emData;
	if(this->params.useSDI_weighing)
	{
		GS_ProblemInput* probInput = this->problem->GetInputBuffer();
		emData = probInput->emData;
		INSIST(probInput->IsEdgeMetricsDefined() == true);
	}

	//CFloatImage emImgs[5];
	//for(int i = 0; i < 5; i++) 
	//{

	//	CShape maskShape = CShape(imgShape.width, imgShape.height, 1);
	//	emImgs[i].ReAllocate(maskShape);
	//	emImgs[i].ClearPixels();
	//}

	const int EquationVars     = 5;
	const int dx[EquationVars] = { 0, -1,  0,  1,  0};
	const int dy[EquationVars] = {-1,  0,  0,  0,  1};

	int levelsToSolve  = (int) this->params.levelsToSolve.size();
	size_v*  eqVarAddr = new size_v[EquationVars   + levelsToSolve];
	precise* eqVarCoef = new precise[EquationVars + levelsToSolve];

	GS_ProblemLevel* levelProb = this->problem->GetLevel(iProbLevel);
	GS_Image levelOrg,  levelFlt;
	GS_Image levelData, levelDataW;
	GS_Image levelOrgGF_X, levelOrgGF_Y;
	GS_Image levelFltGF_X, levelFltGF_Y;

	levelProb->GetData(levelOrg, levelFlt, 
					   levelData, levelDataW,
					   levelOrgGF_X, levelOrgGF_Y, levelFltGF_X, levelFltGF_Y);

	GS_Image dataImg, dataImgW;
	if(this->params.useGlobalDatacost)
	{
		dataImg  = this->problem->GetInputBuffer()->data;
		dataImgW = this->problem->GetInputBuffer()->dataW;
	}
	else
	{
		dataImg  = levelData;
		dataImgW = levelDataW;
	}	
	
	uint nodeAddr  = iSolLevel * imgShape.NodeCount();
	uint pixelAddr = iChannel;
	for(int y = 0; y < imgShape.height; y++)
	for(int x = 0; x < imgShape.width;  x++, nodeAddr++, pixelAddr += imgShape.nBands)
	{
		double rhs = 0;
		double fxy = levelFlt[pixelAddr];

		int    eqVarCount   = 0;
		int    nodeVarIndex = -1;
		double nodeVarCoef  = 0.0;
		for(int iEqVar = 0; iEqVar < EquationVars; iEqVar++)
		{
			if(iEqVar == 2)
			{
				INSIST(dx[iEqVar] == 0);
				INSIST(dy[iEqVar] == 0);

				eqVarAddr[eqVarCount] = nodeAddr;
				nodeVarIndex = eqVarCount;
				eqVarCount++;

				nodeVarCoef += this->params.fidelity_data * dataImgW[pixelAddr];
				rhs         += this->params.fidelity_data * dataImgW[pixelAddr] * dataImg[pixelAddr];		

				//emImgs[iEqVar][nodeAddr] = (float) this->params.fidelity_data * dataImgW[pixelAddr];
				continue;
			}

			int neighbourX = x + dx[iEqVar];
			int neighbourY = y + dy[iEqVar];
			if(imgShape.InBounds(neighbourX, neighbourY) == false) continue;
	
			uint neighbourNodeAddr  = imgShape.NodeIndex(neighbourX, neighbourY);
			uint neighbourPixelAddr = (neighbourNodeAddr * imgShape.nBands) + iChannel;
			neighbourNodeAddr      += iSolLevel * imgShape.NodeCount();
			double neighPinch;
			double neighGrad;
			double neighCoef;

			if(neighbourY != y)
			{
				INSIST(neighbourX == x);
				if(y > neighbourY)
				{
					neighGrad  = (double) levelFltGF_Y[pixelAddr];
					neighPinch = (double) myAbs(levelFltGF_Y[pixelAddr] - levelOrgGF_Y[pixelAddr]);
				}
				else
				{
					neighGrad  = (double) -levelFltGF_Y[neighbourPixelAddr];
					neighPinch = (double) myAbs(levelFltGF_Y[neighbourPixelAddr] - levelOrgGF_Y[neighbourPixelAddr]);
				}
			}
			else
			{
				INSIST(neighbourX != x);
				if(x > neighbourX)
				{
					neighGrad  = (double) levelFltGF_X[pixelAddr];
					neighPinch = (double) myAbs(levelFltGF_X[pixelAddr] - levelOrgGF_X[pixelAddr]);
				}
				else
				{
					neighGrad  = (double) -levelFltGF_X[neighbourPixelAddr];
					neighPinch = (double) myAbs(levelFltGF_X[neighbourPixelAddr] - levelOrgGF_X[neighbourPixelAddr]);
				}		
			}

			neighCoef = this->params.fidelity_sGrad / pow((this->params.pinchFuncScale * neighPinch) + 1.0, this->params.pinchFuncExp);
			
			//bug
			if(this->params.useSDI_weighing)
			{
				double emWeight, emLen;
				if(neighbourY != y)
				{
					if(y > neighbourY)
					{
						emWeight = pow(sin(emData.edgeDirc[nodeAddr]), 2.0f);
						emLen    = emData.edgeLen[nodeAddr];

					}
					else
					{
						emWeight = pow(sin(emData.edgeDirc[neighbourNodeAddr]), 2.0f);
						emLen    = emData.edgeLen[neighbourNodeAddr];
					}
				}
				else
				{
					if(x > neighbourX)
					{
						emWeight = pow(cos(emData.edgeDirc[nodeAddr]), 2.0f);
						emLen    = emData.edgeLen[nodeAddr];

					}
					else
					{
						emWeight = pow(cos(emData.edgeDirc[neighbourNodeAddr]), 2.0f);
						emLen    = emData.edgeLen[neighbourNodeAddr];
					}			
				}

				emWeight *= emLen / this->params.sdiNorm;
				neighCoef /= (1.0 + emWeight);
			}

			//emImgs[iEqVar][nodeAddr] = (float) neighCoef;

			if(neighCoef != 0.0f)
			{
				if(neighbourNodeAddr >= nodeAddr)
				{
					eqVarAddr[eqVarCount] = neighbourNodeAddr;
					eqVarCoef[eqVarCount] = precise(-neighCoef);
					eqVarCount++;
				}

				rhs += neighCoef * neighGrad;
				nodeVarCoef += neighCoef;
			}
		}

		if(this->params.useGlobalDatacost)
		{
			for(int vNeigh = iSolLevel + 1; vNeigh < levelsToSolve; vNeigh++)
			{
				eqVarAddr[eqVarCount] = nodeAddr + ((vNeigh - iSolLevel) * imgShape.NodeCount());
				eqVarCoef[eqVarCount] = (precise) (this->params.fidelity_data * dataImgW[pixelAddr]);
				eqVarCount++;
			}
		}
		
		eqVarCoef[nodeVarIndex] = (precise) nodeVarCoef;
		ENSURE(nodeVarCoef > 0.0);
		solver.SetRow(nodeAddr, eqVarCount, precise(fxy), precise(rhs), eqVarAddr, eqVarCoef);
	}

	delete [] eqVarAddr;
	delete [] eqVarCoef;
}

