#include "GS_VSolver.h"
#include "ImageProcessing.h"

void GS_VSolver::collectSpatialNeighbours(VarCoefList* neighsInFrame)
{
	const int   neighourCount       = 5;
	const int   dx[neighourCount]   = { 0,    -1,    0,    1,    0};
	const int   dy[neighourCount]   = {-1,     0,    0,    0,    1};
    float coef[neighourCount]       = {1.0f, 1.0f, 0.0f, 1.0f, 1.0f};

	for(int iNeighbour = 0; iNeighbour < neighourCount; iNeighbour++)
	{
		coef[iNeighbour] *= (float) this->params.fidelity_sGrad;
	}

	uint nodeAddr  = 0;
	for(int y = 0; y < this->maskShape.height; y++)
	for(int x = 0; x < this->maskShape.width;  x++, nodeAddr++)
	{
		VarCoefList& currNodeNeighbours = neighsInFrame[nodeAddr];
		for(int iNeighbour = 0; iNeighbour < neighourCount; iNeighbour++)
		{
			int neighbourX = x + dx[iNeighbour];
			int neighbourY = y + dy[iNeighbour];
			if(this->maskShape.InBounds(neighbourX, neighbourY) == true)
			{
				uint neighbourNodeAddr = this->maskShape.NodeIndex(neighbourX, neighbourY);
				VarCoef varCoefPair(neighbourNodeAddr, coef[iNeighbour]);
				VarInsertResult result = currNodeNeighbours.insert(varCoefPair);
				ENSURE(result.second == true);
			}
		}
	}
}


void GS_VSolver::collectTemporalNeighbours(VarCoefList* neighsInFrame, 
										   const CFloatImage& outTemporalFlow, bool collectOutNeighbours,
										   const CFloatImage& inTemporalFlow,  bool collectInNeighbours)
{
	INSIST(this->flowShape.nBands == 3);
	if(collectOutNeighbours)
	{
		INSIST(outTemporalFlow.Shape() == this->flowShape);
	}
	if(collectInNeighbours)
	{
		INSIST(inTemporalFlow.Shape()  == this->flowShape);
	}

	INSIST((collectOutNeighbours == true) || (collectInNeighbours == true));
	float temporalCoef = 1.0f * (float) this->params.fidelity_tGrad;
	
	uint nodeAddr = 0;
	uint flowAddr = 0;
	for(int flowType = 0; flowType <= 1; flowType++)
	{
		CFloatImage currFlow;

		if(flowType == 0)
		{
			if (collectOutNeighbours == false) continue;
		}
		else
		{
			ENSURE(flowType == 1);
			if (collectInNeighbours == false) continue;
		}

		if(params.useFlow)
		{
			currFlow = (flowType == 0) ? outTemporalFlow.Clone() : inTemporalFlow.Clone();
		}
		else
		{
			currFlow.ReAllocate(this->flowShape);
			currFlow.ClearPixels();
			ImageProcessing::Operate(currFlow, currFlow, 1.0f, ImageProcessing::OP_SET, 2);
		}
		if(params.useFlowCnf == false)
		{
			ImageProcessing::Operate(currFlow, currFlow, 1.0f, ImageProcessing::OP_SET, 2);			
		}


		for(int y = 0; y < this->flowShape.height; y++)
		for(int x = 0; x < this->flowShape.width;  x++, nodeAddr++, flowAddr += this->flowShape.nBands)
		{
			float u   = currFlow[flowAddr + 0];
			float v   = currFlow[flowAddr + 1];
			float cnf = currFlow[flowAddr + 2];

			//check - black vs sand
			float targetX = x + u;
			float targetY = y + v;

			if(cnf < 1.0f)
				cnf = pow(cnf, (float) this->params.flowCnfPow);
			float flowVecCoef = (temporalCoef * cnf);

			if(this->maskShape.InBounds(targetX, targetY))
			{
				int floorTargetX = (int) targetX;
				int roofTargetX  = (int) (targetX + 1.0f);

				int floorTargetY = (int) targetY;
				int roofTargetY  = (int) (targetY + 1.0f);

				float xRoofFrac  = targetX - floorTargetX;
		  		float yRoofFrac  = targetY - floorTargetY;
				float xFloorFrac = 1.0f - xRoofFrac;
				float yFloorFrac = 1.0f - yRoofFrac;

				for(int iNeigh = 0; iNeigh < 4; iNeigh++)
				{
					int neighX = INT_MAX, neighY = INT_MAX;
					float neighCoefFrac = 0.0f;

					switch(iNeigh)
					{
					case 0:
						neighX = floorTargetX; neighY = floorTargetY;
						neighCoefFrac = xFloorFrac * yFloorFrac;
						break;
					case 1:
						neighX = floorTargetX; neighY = roofTargetY;
						neighCoefFrac = xFloorFrac * yRoofFrac;
						break;
					case 2:
						neighX = roofTargetX;  neighY = floorTargetY;
						neighCoefFrac = xRoofFrac * yFloorFrac;
						break;
					case 3:
						neighX = roofTargetX;  neighY = roofTargetY;
						neighCoefFrac = xRoofFrac * yRoofFrac;
						break;

					default:
						REPORT_FAILURE("Unknown neigh type");
						break;
					}

					if(neighCoefFrac > FLT_EPSILON)
					{
						ENSURE(this->maskShape.InBounds(neighX, neighY));
						float neighCoef = neighCoefFrac * flowVecCoef;

						uint targetAddr = this->maskShape.NodeIndex(neighX, neighY);
						
						VarInsertResult insertResult;
						if(flowType == 0)
						{
							VarCoef varCoefPair(targetAddr, neighCoef);
							insertResult = neighsInFrame[nodeAddr].insert(varCoefPair);
						}
						else
						{
							VarCoef varCoefPair(nodeAddr, neighCoef);
							insertResult = neighsInFrame[targetAddr].insert(varCoefPair);
						}

						if(insertResult.second == false)
						{
							insertResult.first->second += neighCoef;
						}
					}
				}
			}
		}
	}
}



//void GS_VSolver::collectTemporalNeighbours(VarCoefList* neighsInFrame, 
//										    CFloatImage &outTemporalFlow, bool collectOutNeighbours,
//										    CFloatImage &inTemporalFlow,  bool collectInNeighbours)
//{
//	INSIST(this->flowShape.nBands == 3);
//	if(collectOutNeighbours)
//	{
//		INSIST(outTemporalFlow.Shape() == this->flowShape);
//	}
//	if(collect
//	INSIST(inTemporalFlow.Shape()  == this->flowShape);
//
//	INSIST((collectOutNeighbours == true) || (collectInNeighbours == true));
//	float temporalCoef = 1.0f * (float) this->params.fidelity_tGrad;
//
//	CFloatImage outFlow;
//	CFloatImage inFlow;
//
//	if(params.useFlow)
//	{
//		outFlow = outTemporalFlow.Clone();
//		inFlow  = inTemporalFlow.Clone();
//	}
//	else
//	{
//		outFlow.ReAllocate(this->flowShape);
//		inFlow.ReAllocate(this->flowShape);
//		outFlow.ClearPixels();
//		inFlow.ClearPixels();
//		ImageProcessing::Operate(outFlow, outFlow, 1.0f, ImageProcessing::OP_SET, 2);
//		ImageProcessing::Operate(inFlow,  inFlow,  1.0f, ImageProcessing::OP_SET, 2);
//	}
//
//	if(params.useFlowCnf == false)
//	{
//		ImageProcessing::Operate(outFlow, outFlow, 1.0f, ImageProcessing::OP_SET, 2);
//		ImageProcessing::Operate(inFlow,  inFlow,  1.0f, ImageProcessing::OP_SET, 2);
//	}
//	
//	uint nodeAddr  = 0;
//	uint flowAddr  = 0;
//	for(int y = 0; y < this->flowShape.height; y++)
//	for(int x = 0; x < this->flowShape.width;  x++, nodeAddr++, flowAddr += this->flowShape.nBands)
//	for(int flowType = 0; flowType <= 1; flowType++)
//	{
//		if(flowType == 0)
//		{
//			if (collectOutNeighbours == false) continue;
//		}
//		else
//		{
//			ENSURE(flowType == 1);
//			if (collectInNeighbours == false) continue;
//		}
//
//		float u;
//		float v;  
//		float cnf;
//
//		if(flowType == 0)
//		{
//			u    = outFlow[flowAddr + 0];
//			v    = outFlow[flowAddr + 1];
//			cnf  = outFlow[flowAddr + 2];
//		}
//		else
//		{
//			u    = inFlow[flowAddr + 0];
//			v    = inFlow[flowAddr + 1];
//			cnf  = inFlow[flowAddr + 2];
//		}
//
//		//check - black vs sand
//		float targetX = x + u;
//		float targetY = y + v;
//
//		if(cnf < 1.0f)
//			cnf = pow(cnf, (float) this->params.flowCnfPow);
//		float flowVecCoef = (temporalCoef * cnf);
//
//		if(this->maskShape.InBounds(targetX, targetY))
//		{
//			int floorTargetX = (int) targetX;
//			int roofTargetX  = (int) (targetX + 1.0f);
//
//			int floorTargetY = (int) targetY;
//			int roofTargetY  = (int) (targetY + 1.0f);
//
//			float xRoofFrac  = targetX - floorTargetX;
//		  	float yRoofFrac  = targetY - floorTargetY;
//			float xFloorFrac = 1.0f - xRoofFrac;
//			float yFloorFrac = 1.0f - yRoofFrac;
//
//			for(int iNeigh = 0; iNeigh < 4; iNeigh++)
//			{
//				int neighX = INT_MAX, neighY = INT_MAX;
//				float neighCoefFrac = 0.0f;
//
//				switch(iNeigh)
//				{
//				case 0:
//					neighX = floorTargetX; neighY = floorTargetY;
//					neighCoefFrac = xFloorFrac * yFloorFrac;
//					break;
//				case 1:
//					neighX = floorTargetX; neighY = roofTargetY;
//					neighCoefFrac = xFloorFrac * yRoofFrac;
//					break;
//				case 2:
//					neighX = roofTargetX;  neighY = floorTargetY;
//					neighCoefFrac = xRoofFrac * yFloorFrac;
//					break;
//				case 3:
//					neighX = roofTargetX;  neighY = roofTargetY;
//					neighCoefFrac = xRoofFrac * yRoofFrac;
//					break;
//
//				default:
//					REPORT_FAILURE("Unknown neigh type");
//					break;
//				}
//
//				if(neighCoefFrac > FLT_EPSILON)
//				{
//					ENSURE(this->maskShape.InBounds(neighX, neighY));
//					float neighCoef = neighCoefFrac * flowVecCoef;
//
//					uint targetAddr = this->maskShape.NodeIndex(neighX, neighY);
//					
//					VarInsertResult insertResult;
//					if(flowType == 0)
//					{
//						VarCoef varCoefPair(targetAddr, neighCoef);
//						insertResult = neighsInFrame[nodeAddr].insert(varCoefPair);
//					}
//					else
//					{
//						VarCoef varCoefPair(nodeAddr, neighCoef);
//						insertResult = neighsInFrame[targetAddr].insert(varCoefPair);
//					}
//
//					if(insertResult.second == false)
//					{
//						insertResult.first->second += neighCoef;
//					}
//				}
//			}
//		}
//	}
//}
