#include "LinearSolver.h"
#include "SparseSqrSymMat.h"
#include "Timer.h"

//Solve linear system using conjugate gradients
//Algorithm follows the B2 outline found in www.cs.cmu.edu/~quake-papers/painless-conjugate-gradient.pdf 
LinearSolRet 
LinearSolver::Solve_CG_pro(size_v n, AbstractMat *A, precise *x, precise *b, 
  						   Params params)
{
	INSIST(A->GetRowCount() == n);
	INSIST(A->GetColCount() == n);

	precise *r    = new precise[n]; //r = residual vector	
	precise *d    = new precise[n]; //d = conjugate direction	
	precise *Av   = new precise[n]; //Av holds the vec result from A * v (where v = r, x, or d)	
	precise rtr;                    //rtr = r' * r	
	precise rtrOrg;                 //rtrOrg = rtr at the first iteration
	int iter;                       //iterations
	precise epsilonSqr;             //error tolerance squared            
	precise rtrMax;                 //the maximum allowable residual
	precise alpha;                  //length to travel in the direction d 

	INSIST(r  != NULL);
	INSIST(Av != NULL);

	//r = b - Ax
	A->Multiply(n, n, x, Av);   // Av  = A * x
	//#pragma omp parallel for
	for(size_v i = 0; i < n; i++)
	{
		r[i] = b[i] - Av[i];
	}

	//d = r
	Vec::Copy(n, d, r);
	
	iter       = 0;
	rtr        = Vec::Multiply(n, r, r);
	rtrOrg     = rtr;	

	//epsilonSqr = params.errTolerance * params.errTolerance;
	//rtrMax     = rtrOrg * epsilonSqr;
	epsilonSqr = params.errTolerance;
	rtrMax     = epsilonSqr;
	 
	while((iter < params.maxIters) &&
		  (rtr > rtrMax))
	{
		A->Multiply(n, n, d, Av);                      //Av = A * d
		precise alphaDenom = Vec::Multiply(n, d, Av);  //alphaDenom = d' * (A * d)

		if(alphaDenom != 0.0f)
			alpha = rtr / alphaDenom;  //alpha = rtr / (d' * (A * d))
		else // d is the null vector
		{			
			break;
		}

		//x = x + ad
		//#pragma omp parallel for 
		for(size_v i = 0; i < n; i++)
		{
			x[i] += alpha * d[i];
		}

		if((iter % params.fltErrMaxiters == 0) &&
		   (iter > 0)) 
		{
			//r = b - Ax
			A->Multiply(n, n, x, Av);  
			//#pragma omp parallel for 
			for(size_v i = 0; i < n; i++)
			{
				r[i] = b[i] - Av[i];
			}
		}
		else
		{
			//r = r - (alpha * (A * r))
			//#pragma omp parallel for 
			for(size_v i = 0; i < n; i++)
			{
				r[i] -= alpha * Av[i];
			}
		}


		precise rtrOld = rtr;
		rtr = Vec::Multiply(n, r, r);
		precise beta = rtr / rtrOld;

		//d = r + beta*d
		//#pragma omp parallel for 
		for(size_v i = 0; i < n; i++)
		{
			d[i] = r[i] + (beta * d[i]);
		}

		iter++;
	}

	delete [] r;
	delete [] d;
	delete [] Av;

	return LinearSolRet(rtr, iter);
}

//Solve linear system using gradient descent
//Algorithm follows the B3 outline found in www.cs.cmu.edu/~quake-papers/painless-conjugate-gradient.pdf 
LinearSolRet 
LinearSolver::Solve_PCG_pro(size_v n, AbstractMat* A, precise* x, precise* b, 
							Params params, AbsractPreconditioner* MInv)
{
	INSIST(A->GetRowCount() == n);
	INSIST(A->GetColCount() == n);
	
	precise* r  = new precise[n];   //r = residual vector	
	precise* d  = new precise[n];   //d = conjugate direction	
	precise* s  = new precise[n];   //s = preconditioned r
	precise* Av = new precise[n];   //Av holds the vec result from A * v (where v = r, x, or d)	
	precise rtr;                    //rtr = r' * r	
	precise rtrOrg;                 //rtrOrg = rtr at the first iteration
	int iter;                       //iterations
	precise epsilonSqr;             //error tolerance squared            
	precise rtrMax;                 //the maximum allowable residual
	precise alpha;                  //length to travel in the direction d 

	INSIST(r  != NULL);
	INSIST(Av != NULL);


	//r = b - Ax
	A->Multiply(n, n, x, Av);   // Av  = A * x

	//#pragma omp parallel for
	for(size_v i = 0; i < n; i++)
	{
		r[i] = b[i] - Av[i];
	}

	//d = MInv * r
	Vec::Copy(n, d, r);
	if(MInv != NULL)
		MInv->Apply(n, d);
	
	iter       = 0;
	rtr        = Vec::Multiply(n, r, d);
	rtrOrg     = rtr;	

	//epsilonSqr = params.errTolerance * params.errTolerance;
	//rtrMax     = rtrOrg * epsilonSqr;
	epsilonSqr = params.errTolerance;
	rtrMax     = epsilonSqr;

	while((iter < params.maxIters) &&
		  (rtr > rtrMax))
	{
		A->Multiply(n, n, d, Av);                      //Av = A * d
		precise alphaDenom = Vec::Multiply(n, d, Av);  //alphaDenom = d' * (A * d)

		if(alphaDenom != 0.0f)
			alpha = rtr / alphaDenom;  //alpha = rtr / (d' * (A * d))
		else // d is the null vector
		{			
			break;
		}

		//x = x + ad
		//#pragma omp parallel for
		for(size_v i = 0; i < n; i++)
		{
			x[i] += alpha * d[i];
		}

		if((iter % params.fltErrMaxiters == 0) &&
		   (iter > 0)) 
		{			
			//r = b - Ax
			A->Multiply(n, n, x, Av);  
			//#pragma omp parallel for
			for(size_v i = 0; i < n; i++)
			{
				r[i] = b[i] - Av[i];
			}
		}
		else
		{			
			//r = r - (alpha * (A * r))
			//#pragma omp parallel for
			for(size_v i = 0; i < n; i++)
			{
				r[i] -= alpha * Av[i];
			}
		}

		//s = Minv * r
		Vec::Copy(n, s, r);
		if(MInv != NULL)
			MInv->Apply(n, s);

		precise rtrOld = rtr;
		rtr = Vec::Multiply(n, r, s);
		precise beta = rtr / rtrOld;

		//d = s + beta*d
		//#pragma omp parallel for 
		for(size_v i = 0; i < n; i++)
		{
			d[i] = s[i] + (beta * d[i]);
		}

		iter++;
	}

	delete [] r;
	delete [] d;
	delete [] s;
	delete [] Av;

	return LinearSolRet(rtr, iter);
}

