#include "GSF_ELE.h"
#include "ImageIO.h"
#include "Disk.h"


GS_Problem::Params GSF_ELE::GS_Problem_Params_Default()
{
	GS_Problem::Params params = GS_Problem::Params::Default();
	params.plsParams.levelCount = 1;
	params.channelsType = CT_RGB;
	return params;
}

GS_Problem::Params GSF_ELE::GS_Problem_Params_MultiFreq()
{
	GS_Problem::Params params = GS_Problem::Params::Default();
	params.plsParams.levelCount  = 3;
	params.plsParams.sigmaD      = 10.0f;
	params.plsParams.sigmaD_mult = 1.0f;
	params.plsParams.levelsType  = GS_ProblemLevels::LT_DOG;
	return params;
}


GS_Solver::Params GSF_ELE::GS_Solver_Params_Default()
{
	GS_Solver::Params params = GS_Solver::Params::Default();
	params.fidelity_data = 3.0E-02;
	return params;
}

GS_Image GSF_ELE::ELE(GS_ProblemInput& probInput,
					  Params fltParams,
					  GS_Problem::Params probParams,
					  GS_Solver::Params solParams)
{
	INCR_TAB_LEVEL();
	string paramsSpacing((TAB_SPACE_COUNT + 1) * TAB_LEVEL, ' ');
	
	REPORT("Creating GS problem... \n");
	probParams.Dump(paramsSpacing);
	GS_Problem prob(probInput, probParams);	

	REPORT("Filtering GS problem... \n");
	fltParams.Dump(paramsSpacing);
	GSF_ELE eleFlt(&prob, fltParams);
	if(solParams.levelsToSolve.size() == 0)
		eleFlt.Filter();
	else
		eleFlt.Filter(solParams.levelsToSolve);

	REPORT("Solving GS problem... \n");
	solParams.Dump(paramsSpacing);
	GS_Solver solver(&prob, solParams);
	solver.Solve();

	DCR_TAB_LEVEL();
	return prob.GetResultRGB();	
}


GS_Image GSF_ELE::ELE_Freq(GS_ProblemInput& probInput, int iFreqStart, int iFreqEnd,
						   Params fltParams,
						   GS_Problem::Params probParams,
						   GS_Solver::Params solParams)
{
	INSIST(iFreqStart <= iFreqEnd);
	INSIST(iFreqStart >= 0);
	INSIST(iFreqEnd < probParams.plsParams.levelCount);

	INSIST(solParams.levelsToSolve.size() == 0);
	for(int iFreq = iFreqStart; iFreq <= iFreqEnd; iFreq++)
	{
		solParams.levelsToSolve.insert(iFreq);
	}

	return ELE(probInput, fltParams, probParams, solParams);
}

GS_Image GSF_ELE::ELE_HiFreq(GS_ProblemInput& probInput,
				             Params fltParams,
				             GS_Problem::Params probParams,
				             GS_Solver::Params solParams)
{
	int iFreqStart = NEAREST_INT((0.0f / 3.0f) * probParams.plsParams.levelCount);
	int iFreqEnd   = NEAREST_INT((1.0f / 3.0f) * probParams.plsParams.levelCount);
	if(iFreqEnd > iFreqStart)
		iFreqEnd--;
	return ELE_Freq(probInput, iFreqStart, iFreqEnd,
			        fltParams, probParams, solParams);
}

GS_Image GSF_ELE::ELE_MidFreq(GS_ProblemInput& probInput,
							  Params fltParams,
							  GS_Problem::Params probParams,
							  GS_Solver::Params solParams)
{
	int iFreqStart = NEAREST_INT((1.0f / 3.0f) * probParams.plsParams.levelCount);
	int iFreqEnd   = NEAREST_INT((2.0f / 3.0f) * probParams.plsParams.levelCount);
	if(iFreqEnd > iFreqStart)
		iFreqEnd--;
	return ELE_Freq(probInput, iFreqStart, iFreqEnd,
			        fltParams, probParams, solParams);
}

GS_Image GSF_ELE::ELE_LowFreq(GS_ProblemInput& probInput,
						      Params fltParams,
						      GS_Problem::Params probParams,
						      GS_Solver::Params solParams)
{
	int iFreqStart = NEAREST_INT((2.0f / 3.0f) * probParams.plsParams.levelCount);
	int iFreqEnd   = NEAREST_INT((3.0f / 3.0f) * probParams.plsParams.levelCount);
	INSIST(iFreqEnd > iFreqStart);
	if(iFreqEnd > iFreqStart)
		iFreqEnd--;
	return ELE_Freq(probInput, iFreqStart, iFreqEnd,
			        fltParams, probParams, solParams);
}



void GSF_ELE::TestFilters(string imgFN, string outDir)
{
	GS_Problem::Params      probParams      = GS_Problem_Params_Default();
	GS_ProblemInput::Params probInputParams = GS_ProblemInput::Params::Default();
	probInputParams.loadEmData = true;
	GS_ProblemInput probInput(imgFN, probInputParams, probParams.channelsType);

	string fileID = Disk::GetFileNameID(Disk::GetBaseFileName(imgFN));
	GS_Image fltImg;

	Params defParams    = GSF_ELE::Params::Default();
	Params defInvParams = GSF_ELE::Params::DefaultInv();
	
	//ele tests
	fltImg = GSF_ELE::ELE(probInput, defParams);
	ImageIO::WriteFile(fltImg, outDir + fileID + "-ele-default.tga");

	//fltImg = GSF_ELE::ELE(probInput, defInvParams);
	//ImageIO::WriteFile(fltImg, outDir + fileID + "-ele-defaultInv.tga");

	//fltImg = GSF_ELE::ELE_HiFreq(probInput, defParams);
	//ImageIO::WriteFile(fltImg, outDir + fileID + "-ele-hiFreq.tga");

	//fltImg = GSF_ELE::ELE_LowFreq(probInput, defParams);
	//ImageIO::WriteFile(fltImg, outDir + fileID + "-ele-lowFreq.tga");
}

