#include "EdgeMetrics.h"
#include "ImageProcessing.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "ImageIO.h"
#include "Disk.h"

EdgeMetrics EdgeMetrics::Clone() const
{
	EdgeMetrics clone;
	clone.edgeDirc     = this->edgeDirc.Clone();
	clone.edgeLen      = this->edgeLen.Clone();
	clone.edgeStrength = this->edgeStrength.Clone();
	return clone;
}

void EdgeMetrics::Load(string fileName, CShape imgShape)
{
	FILE *inFile = fopen(fileName.c_str(), "rb");
	INSIST(inFile != NULL);

	CShape maskShape(imgShape.width, imgShape.height, 1);
	this->edgeDirc.ReAllocate(maskShape);
	this->edgeLen.ReAllocate(maskShape);
	this->edgeStrength.ReAllocate(maskShape);

	for(int iStat = 0; iStat < 3; iStat++)
	{
		char *memBuff = NULL;
		switch(iStat)
		{
		case 0:
			memBuff = (char *) this->edgeStrength.PixelAddress(0, 0, 0);
			break;
		case 1:
			memBuff = (char *) this->edgeDirc.PixelAddress(0, 0, 0);
			break;
		case 2:
			memBuff = (char *) this->edgeLen.PixelAddress(0, 0, 0);
			break;
		default:
			REPORT_FAILURE("Unknown stat - %\n", iStat);
			break;
		}

		int bytesToRead = maskShape.width * maskShape.height * sizeof(float);
		int totalBytesRead = 0;
		while(bytesToRead != 0)
		{
			int freadCount;
			#ifdef WIN32
				freadCount = min(bytesToRead, 64000);
			#else
				freadCount = bytesToRead;
			#endif

			int bytesRead = (int) fread((void *) (memBuff + totalBytesRead), 
											sizeof(*memBuff),
											freadCount,
											inFile);

			totalBytesRead += bytesRead;
			bytesToRead -= bytesRead;

			if(bytesRead == 0)
			{
				int errorCode = ferror(inFile);
				if(errorCode != 0)
				{
					PRINT_INT(errorCode);
					PRINT_INT(errno);
					PRINT_STR(strerror(errno));
					REPORT_FAILURE("IO failure");
				}
			}	
		}
	}

	this->edgeStrength = ImageProcessing::FlipVertical(this->edgeStrength);
	this->edgeLen      = ImageProcessing::FlipVertical(this->edgeLen);
	this->edgeDirc     = ImageProcessing::FlipVertical(this->edgeDirc);


	//bug - remove once you switch to your own edge detector
	correctLarryEdgeDirc();

	//Dump();

	fclose(inFile);
}




void EdgeMetrics::Dump()
{
	CShape maskShape = this->edgeDirc.Shape();
	 
	PRINT_IMG(this->edgeStrength);
	ImageProcessing::Operate(this->edgeStrength, this->edgeStrength, 255.0f, ImageProcessing::OP_DIV);

	PRINT_IMG(this->edgeDirc);
	ImageProcessing::Operate(this->edgeDirc, this->edgeDirc, 3.1415f, ImageProcessing::OP_DIV);
	
	PRINT_IMG(this->edgeLen);
	ImageProcessing::Operate(this->edgeLen, this->edgeLen, 1500.0f, ImageProcessing::OP_DIV);


	ImageIO::WriteFile(this->edgeStrength, "1.tga");
	ImageIO::WriteFile(this->edgeDirc, "2.tga");
	ImageIO::WriteFile(this->edgeLen, "3.tga");
	exit(0);
}


string EdgeMetrics::GetFileName(string imgFN)
{
	return Disk::GetFileNameID(imgFN) + ".edg";
}

void EdgeMetrics::correctLarryEdgeDirc()
{
	CShape emShape = this->edgeDirc.Shape();
	INSIST(emShape.nBands == 1);
	int nodeCount = emShape.width * emShape.height * emShape.nBands;

	float radShift = MathHelper::DegressToRadians(90.0f);
	float M_2PI    = 2.0f * M_PI_FLOAT;

	for(int iNode = 0; iNode < nodeCount; iNode++)
	{
		float dirc = this->edgeDirc[iNode];

		CHECK_RANGE(dirc, -FLT_EPSILON, M_2PI);

		//rotate by 90 degree
		dirc += radShift;
		if(dirc > M_2PI)
		{
			dirc -= M_2PI;
		}

		//reverse direction
		dirc = M_2PI - dirc;

		this->edgeDirc[iNode] = dirc;
	}
}
