#include "FlowHelper.h"
#include "ImageProcessing.h"
#include <float.h>
#include "ImageIO.h"
#include "Disk.h"

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

void FlowHelper::WriteFlowField(const char *fname,
							   CFloatImage &flowField)
{
	CShape fieldShape = flowField.Shape();	
	ENSURE(fieldShape.depth	== 1);
	ENSURE(fieldShape.nBands == 3);

	uint frameSize    = fieldShape.width * fieldShape.height * fieldShape.nBands;
	float *tempBuffer = new float[frameSize]; 	
	ENSURE(tempBuffer != NULL);

	for(uint iElem = 0; iElem < frameSize; iElem++)
	{
		tempBuffer[iElem] = flowField[iElem];
	}

	FILE *filePtr = fopen(fname, "wb");
	if(filePtr == NULL)
	{
		REPORT_FAILURE("Error writing flow file - %s\n", fname);
	}
	if((fwrite(&fieldShape.width,  sizeof(fieldShape.width),  1, filePtr) != 1) ||
	   (fwrite(&fieldShape.height, sizeof(fieldShape.height), 1, filePtr) != 1) ||
	   (fwrite(&fieldShape.nBands, sizeof(fieldShape.nBands), 1, filePtr) != 1))
	{
		REPORT_FAILURE("Error writing flow file");
	}
	if(fwrite(tempBuffer, sizeof(*tempBuffer), frameSize, filePtr) != frameSize)
	{
		REPORT_FAILURE("Error writing flow file");
	}
	fclose(filePtr);

	delete tempBuffer;

}

void FlowHelper::ReadFlowField(const char *fname,
							   CFloatImage &flowField,
							   float scale)
{
	FILE *filePtr = fopen(fname, "rb");
	if(filePtr == NULL)
	{
		REPORT_FAILURE("Couldn't open file - %s\n", fname);
	}

	CShape fieldShape;
	fieldShape.depth  = 1;
	if((fread(&fieldShape.width,  sizeof(fieldShape.width),  1, filePtr) != 1) ||
	   (fread(&fieldShape.height, sizeof(fieldShape.height), 1, filePtr) != 1) ||
	   (fread(&fieldShape.nBands, sizeof(fieldShape.nBands), 1, filePtr) != 1))
	{
		REPORT_FAILURE("Error reading flow file");
	}
	ENSURE(fieldShape.nBands == 3);

	uint frameSize    = fieldShape.width * fieldShape.height * fieldShape.nBands;
	float *tempBuffer = new float[frameSize]; 
	ENSURE(tempBuffer != NULL);

	if(fread(tempBuffer, sizeof(*tempBuffer), frameSize, filePtr) != frameSize)
	{
		REPORT_FAILURE("Error reading flow file - %s", fname);
	}
	fclose(filePtr);

	flowField.ReAllocate(fieldShape);
	for (uint iElem = 0; iElem < frameSize; iElem++)
	{
		flowField[iElem] = tempBuffer[iElem];
	}

	delete tempBuffer;

	if(scale != 1.0f)
	{
		ImageProcessing::Operate(flowField, flowField, scale, ImageProcessing::OP_MUL);
		CFloatImage flowFieldCopy = flowField.Clone();
		ImageProcessing::ScaleImage(flowField, flowFieldCopy, scale);
	}
}

void FlowHelper::ReadSandFlowField(const char *fname, 
								   CFloatImage &flowField,
								   float scale)
{
	FILE *filePtr = fopen(fname, "rb");
	if(filePtr == NULL)
	{
		REPORT_FAILURE("Couldn't open file - %s\n", fname);
	}

	CShape fieldShape;
	fieldShape.depth  = 1;
	fieldShape.nBands = 3;
	if((fread(&fieldShape.width,  sizeof(fieldShape.width),  1, filePtr) != 1) ||
	   (fread(&fieldShape.height, sizeof(fieldShape.height), 1, filePtr) != 1))
	{
		REPORT_FAILURE("Error reading flow file");
	}

	uint nodeCount = fieldShape.width * fieldShape.height;
	flowField.ReAllocate(fieldShape);
	uint pixelAddr = 0;
	for (uint iElem = 0; iElem < nodeCount; iElem++, pixelAddr += 3)
	{
		float u, v;
		int occ;

	   if((fread(&u,   sizeof(u),   1, filePtr) != 1) ||
	      (fread(&v,   sizeof(v),   1, filePtr) != 1) || 
		  (fread(&occ, sizeof(occ), 1, filePtr) != 1))
	   {
		   REPORT_FAILURE("Error reading flow file - %s", fname);
	   }

	   CHECK_RANGE(occ, 0, 255);
	   float confidence = (255.0f - occ) / 255.0f;

	   flowField[pixelAddr + 0] = u;
	   flowField[pixelAddr + 1] = v;		   
	   flowField[pixelAddr + 2] = confidence;
	}

	if(scale != 1.0f)
	{
		ImageProcessing::Operate(flowField, flowField, scale, ImageProcessing::OP_MUL);
		CFloatImage flowFieldCopy = flowField.Clone();
		ImageProcessing::ScaleImage(flowField, flowFieldCopy, scale);
	}
}


CFloatImage FlowHelper::ColorizeFlowField(const CFloatImage &flowField) 
{
	CShape ffShape = flowField.Shape();
	INSIST(ffShape.nBands == 3);
	CFloatImage resultImg(ffShape);
	resultImg.ClearPixels();

	// compute max motion
	float max = 0.0f;
	uint nodeCount = ffShape.width * ffShape.height;
	uint pixelAddr = 0;
	for (uint iElem = 0; iElem < nodeCount; iElem++, pixelAddr += 3)	
	{
		if(flowField[pixelAddr + 2] > 0.0f)
		{
			float val;

			val = fabs(flowField[pixelAddr + 0]);
			SET_IF_GT(max, val);

			val = fabs(flowField[pixelAddr + 1]);
			SET_IF_GT(max, val);
		}
	}
	INSIST(max != 0.0f);

	pixelAddr = 0;
	for (uint iElem = 0; iElem < nodeCount; iElem++, pixelAddr += 3)
	{
		if(flowField[pixelAddr + 2] > 0.5f)
		{
			float u = 0.5f + (flowField[pixelAddr + 0] / max);
			float v = 0.5f + (flowField[pixelAddr + 1] / max);
			resultImg[pixelAddr + 0] = u;
			resultImg[pixelAddr + 1] = 0.5f;
			resultImg[pixelAddr + 2] = v;
		}
	}

	return resultImg; 
}

string FlowHelper::GetFlowFN(string imgFN)
{
	return Disk::GetFileNameID(imgFN) + ".flow";
}
