///////////////////////////////////////////////////////////////////////////
//
// NAME
//  FileIO.h -- image file input/output
//
// DESCRIPTION
//  Read/write image files, potentially using an interface to an
//  external package.
//
//  You can either pass and empty (unitialized) image to ReadFile,
//  or one you have already allocated (with a specific pixel type).
//  
//  If you don't initialize the image, the type of the returned image
//  (e.g., 1 band vs. 4 band) will be determined by the image file.
//  If you do initialize the image, it will be re-allocated if necessary,
//  and the data will be coerced into the type you specified.
//
// SEE ALSO
//  FileIO.cpp          implementation
//
// Copyright  Richard Szeliski, 2001.  See Copyright.h for more details
//
///////////////////////////////////////////////////////////////////////////

#ifndef __IMAGE_IO_H__
#define __IMAGE_IO_H__

#include "Image.h"
#include "tiff.h"

const extern char *IMAGE_IO_EXT_STRS[4];
const extern char *IMAGE_IO_EXT_TYPE[4];

typedef struct _CompressionInfo
{
	int v_samp_factor[3];
	int h_samp_factor[3];
	int quality;

	_CompressionInfo(int imgQuality = 90, int channels = 3)
	{
		quality = imgQuality;
		v_samp_factor[0] = h_samp_factor[0] = 1;
		
		if(channels == 3)
		{
			v_samp_factor[1] = h_samp_factor[1] = 1;
			v_samp_factor[2] = h_samp_factor[2] = 1;
		}
		else
		{
			INSIST(channels == 1);
			v_samp_factor[1] = h_samp_factor[1] = 0;
			v_samp_factor[2] = h_samp_factor[2] = 0;
		}
	}
} CompressionInfo;

class ImageIO
{

public:

	template <class T>
	static void WriteFileTGA(const CImageOf<T> &img, const char* filename);

	template <class T>
	static void ReadFileTGA(CImageOf<T> &img, const char* filename);

	static void WriteFileTiff(const CByteImage& img, const char* filename, 
							  int compressionType = COMPRESSION_NONE);
	static void WriteFileTiff(const CFloatImage& img, const char* filename,
							  int compressionType = COMPRESSION_NONE);
	static void WriteFileTiff(const CHalfImage& img, const char* filename,
							  int compressionType = COMPRESSION_NONE);

	static void WriteFileTiff(const CFloatImage& img, string filename, 
							  int compressionType = COMPRESSION_NONE)
	{
		WriteFileTiff(img, filename.c_str(), compressionType);
	}
	static void WriteFileTiff(const CByteImage& img, string filename, 
							  int compressionType = COMPRESSION_NONE)
	{
		WriteFileTiff(img, filename.c_str(), compressionType);
	}
	static void WriteFileTiff(const CHalfImage& img, string filename, 
							  int compressionType = COMPRESSION_NONE)
	{
		WriteFileTiff(img, filename.c_str(), compressionType);
	}

	static void ReadFileTiff(CByteImage&  img, const char* filename);
	static void ReadFileTiff(CFloatImage& img, const char* filename);
	static void ReadFileTiff(CHalfImage&  img, const char* filename);


	static void WriteFileJPG(const CByteImage& img, const char* filename,
							 const CompressionInfo &compressionInfo = CompressionInfo());
	static void WriteFileJPG(const CFloatImage& img, const char* filename,
							 const CompressionInfo &compressionInfo = CompressionInfo());
	static void WriteFileJPG(const CHalfImage& img, const char* filename,
							 const CompressionInfo &compressionInfo = CompressionInfo());

	static void WriteFileJPG(const CByteImage& img, string filename, 
							 const CompressionInfo &compressionInfo = CompressionInfo())
	{
		WriteFileJPG(img, filename.c_str(), compressionInfo);
	}
	static void WriteFileJPG(const CFloatImage& img, string filename, 
							 const CompressionInfo &compressionInfo = CompressionInfo())
	{
		WriteFileJPG(img, filename.c_str(), compressionInfo);
	}
	static void WriteFileJPG(const CHalfImage& img, string filename, 
							 const CompressionInfo &compressionInfo = CompressionInfo())
	{
		WriteFileJPG(img, filename.c_str(), compressionInfo);
	}


	static void ReadFileJPG(CFloatImage& img, const char* filename, CompressionInfo* compressionInfo = NULL);
	static void ReadFileJPG(CByteImage&  img, const char* filename, CompressionInfo* compressionInfo = NULL);
	static void ReadFileJPG(CHalfImage&  img, const char* filename, CompressionInfo* compressionInfo = NULL);

	static void ReadFileJPG(CFloatImage& img, string filename, CompressionInfo *compressionInfo = NULL)
	{
		ReadFileJPG(img, filename.c_str(), compressionInfo);
	}
	static void ReadFileJPG(CByteImage& img, string filename,   CompressionInfo *compressionInfo = NULL)
	{
		ReadFileJPG(img, filename.c_str(), compressionInfo);
	}
	static void ReadFileJPG(CHalfImage& img, string filename,      CompressionInfo *compressionInfo = NULL)
	{
		ReadFileJPG(img, filename.c_str(), compressionInfo);
	}


	static void WriteFileRGBE(const CByteImage&  img, const char* filename);
	static void WriteFileRGBE(const CFloatImage& img, const char* filename);
	static void WriteFileRGBE(const CHalfImage&  img, const char* filename);

	static void WriteFileRGBE(const CByteImage& img, string filename)
	{
		WriteFileRGBE(img, filename.c_str());
	}	
	static void WriteFileRGBE(const CFloatImage& img, string filename)
	{
		WriteFileRGBE(img, filename.c_str());
	}
	static void WriteFileRGBE(const CHalfImage& img, string filename)
	{
		WriteFileRGBE(img, filename.c_str());
	}

	static void ReadFileRGBE(CFloatImage &img, const char* filename);
	static void ReadFileRGBE(CByteImage  &img, const char* filename);
	static void ReadFileRGBE(CHalfImage  &img, const char* filename);

	static void ReadFileRGBE(CFloatImage& img, string filename)
	{
		ReadFileRGBE(img, filename.c_str());
	}	
	static void ReadFileRGBE(CByteImage& img, string filename)
	{
		ReadFileRGBE(img, filename.c_str());
	}
	static void ReadFileRGBE(CHalfImage& img, string filename)
	{
		ReadFileRGBE(img, filename.c_str());
	}


	static void ReadFile(CFloatImage& img,  const char* filename, float scale);
	static void ReadFile(CFloatImage& img,  const char* filename);
	static void ReadFile(CByteImage& img,  const char* filename);
	static void ReadFile(CByteImage& img,  const char* filename, float scale);
	
	static void ReadFile(CFloatImage& img, string filename, float scale)
	{
		ReadFile(img, filename.c_str(), scale);
	}
	static void ReadFile(CFloatImage& img, string filename)
	{
		ReadFile(img, filename.c_str());
	}
	static void ReadFile(CByteImage& img, string filename)
	{
		ReadFile(img, filename.c_str());
	}
	static void ReadFile(CByteImage& img, string filename, float scale)
	{
		ReadFile(img, filename.c_str(), scale);
	}

	template<class T>
	static void WriteFile(const CImageOf<T>& img, const char* filename);	

	template<class T>
	static void WriteFile(const CImageOf<T>& img, string filename)
	{
		WriteFile(img, (const char*) filename.c_str());
	}


	static void WriteVideo(CFloatImage &video, const char* filePrefix);
	static void WriteVideo(CFloatImage &video, string filePrefix)
	{
		WriteVideo(video, filePrefix.c_str());
	}


	static CShape GetImageShape(const char *fileName,  float scale = 1.0f);
	static CShape GetImageShape(string fileName, float scale = 1.0f)
	{
		return GetImageShape(fileName.c_str(), scale);
	}		
	static CShape GetImageShapeTiff(const char *fileName);
	static CShape GetImageShapeJpg(const char *fileName);

	template<class T>
	static void ReadRaw(CImageOf<T> &rawImage, 
						const char *fname,
						float scale = 1.0f);

	template<class T>
	static void ReadRaw(CImageOf<T> &rawImage, 
						string fname,
						float scale = 1.0f)
	{
		ReadRaw(rawImage, fname.c_str(), scale);
	}

	template<class T>
	static void WriteRaw(const CImageOf<T> &image,
						 const char *fname);

	template<class T>
	static void WriteRaw(const CImageOf<T> &rawImage,
						 string fname)
	{
		WriteRaw(rawImage, fname.c_str());
	}
};

#include "ImageIO.inl"
#include "ImageIO-TGA.inl"

#endif


