#include "Curve.h"

void Curve::ComputeFuncMap(int _funcSamples)
{
	computeFuncMapFast(_funcSamples);
	//computeFuncMapPrecise(_funcSamples);
}

void Curve::computeFuncMapFast(int _funcSamples)
{
	INSIST(this->size() >= 2);

	if(this->funcSamples != _funcSamples)
	{
		this->funcSamples = _funcSamples;
		this->funcMap = new double[this->funcSamples + 1];
	}

	for(int iSample = 0; iSample <= this->funcSamples; iSample++)
	{
		this->funcMap[iSample] = DBL_MAX;
	}

	for(int iSeg = 0; iSeg < this->size() - 1; iSeg++)
	{
		CurvePoint p1 = this->points[iSeg];
		CurvePoint p2 = this->points[iSeg + 1];

		CurvePoint p0;
		if(iSeg > 0)
			p0 = this->points[iSeg - 1];
		else
			p0 = CurvePoint(0.0, p1.y);

		CurvePoint p3;
		if(iSeg < this->size() - 2)
			p3 = this->points[iSeg + 2];
		else
			p3 = CurvePoint(1.0, p2.y);

		computeSegFuncMap(p0, p1, p2, p3, this->funcSamples);
	}

	computeFuncMapHoles();
}

void Curve::computeFuncMapPrecise(int _funcSamples)
{
	INSIST(this->size() >= 2);

	if(this->funcSamples != _funcSamples)
	{
		this->funcSamples = _funcSamples;
		this->funcMap = new double[this->funcSamples + 1];
	}
	
	for(int iSample = 0; iSample <= this->funcSamples; iSample++)
	{
		double curveResponseX = iSample / (double) this->funcSamples;
		double curveResponseY = GetCurveResponsePrecise(curveResponseX);
		this->funcMap[iSample] = curveResponseY;
	}
}

/*
 * Catmul rom spline computation: http://www.mvps.org/directx/articles/catmull/
 */
void Curve::computeSegFuncMap(const CurvePoint &p0, const CurvePoint &p1, 
					  	      const CurvePoint &p2, const CurvePoint &p3,
						      int segSamples)
{
	double termX0 = 0.5 * (2.0 * p1.x);
	double termX1 = 0.5 * (p2.x - p0.x);
	double termX2 = 0.5 * ((2.0 * p0.x) - (5.0 * p1.x) + (4.0 * p2.x) - p3.x);
	double termX3 = 0.5 * ((3.0 * p1.x) - (3.0 * p2.x) + p3.x - p0.x);

	double termY0 = 0.5 * (2.0 * p1.y);
	double termY1 = 0.5 * (p2.y - p0.y);
	double termY2 = 0.5 * ((2.0 * p0.y) - (5.0 * p1.y) + (4.0 * p2.y) - p3.y);
	double termY3 = 0.5 * ((3.0 * p1.y) - (3.0 * p2.y) + p3.y - p0.y);

	double responseRange = this->maxOut - this->minOut;
	INSIST(responseRange >= 0.0);

	for(int iSegSample = segSamples; iSegSample >= 0; iSegSample--)
	{
		double t = iSegSample / (double) segSamples;
		double t2 = t * t;
		double t3 = t2 * t;

		double segX = termX0 + (termX1 * t) + (termX2 * t2) + (termX3 * t3);
		double segY = termY0 + (termY1 * t) + (termY2 * t2) + (termY3 * t3);

		segX = max( min(segX, 1.0), 0.0 );
		segY = max( min(segY, 1.0), 0.0 );
		segY = (segY * responseRange) + this->minOut;

		int funcIndex = (int) (segX * this->funcSamples);
		this->funcMap[funcIndex] = segY;
	}
}

void Curve::computeFuncMapHoles()
{
	bool prevSampleWasMissing = false;

	double sampleSigRange = 0;
	int    sampleIdxRange = -1;

	double sampleLeft = -1, sampleRight = -1;
	int jSample = -1, kSample = -1;

	for(int iSample = 0; iSample <= this->funcSamples; iSample++)
	{
		if(this->funcMap[iSample] == DBL_MAX)
		{			
			if(prevSampleWasMissing == false)
			{			
				for(jSample = iSample - 1; jSample >= 0; jSample--)
				{
					if(this->funcMap[jSample] != DBL_MAX)
					{
						sampleLeft = this->funcMap[jSample];
						break;
					}
				}
				
				for(kSample = iSample + 1; kSample <= this->funcSamples; kSample++)
				{
					if(this->funcMap[kSample] != DBL_MAX)
					{
						sampleRight = this->funcMap[kSample];
						break;
					}
				}

				if(jSample == -1)
				{
					INSIST(kSample != this->funcSamples + 1);
					sampleLeft = sampleRight;
				}
				else if(kSample == this->funcSamples + 1)
				{
					sampleRight = sampleLeft;
				}

				sampleSigRange = sampleRight - sampleLeft;
				sampleIdxRange = kSample - jSample;
			}

			this->funcMap[iSample] = (((iSample - jSample) / (double) sampleIdxRange) * sampleSigRange) + sampleLeft;
			prevSampleWasMissing = true;
		}
		else
		{
			prevSampleWasMissing = false;
		}
	}
}

double Curve::GetCurveResponsePrecise(double x, double xDistThres) const
{
	CHECK_RANGE(x, 0.0, 1.0);
	INSIST(this->funcSamples > 0);
	INSIST(this->size() >= 2);

	double y = 0;

	if(x <= this->points.front().x)
		y = this->points.front().y;
	else if(x >= this->points.back().x)
		y = this->points.back().y;
	else
	{
		for(int iSeg = 0; iSeg < this->size() - 1; iSeg++)
		{
			if(InRange(x, this->points[iSeg].x, this->points[iSeg + 1].x))
			{
				CurvePoint p1 = this->points[iSeg];
				CurvePoint p2 = this->points[iSeg + 1];

				CurvePoint p0;
				if(iSeg > 0)
					p0 = this->points[iSeg - 1];
				else
					p0 = CurvePoint(0.0, p1.y);

				CurvePoint p3;
				if(iSeg < this->size() - 2)
					p3 = this->points[iSeg + 2];
				else
					p3 = CurvePoint(1.0, p2.y);

				y = getCurveResponsePrecise(p0, p1, p2, p3, x, xDistThres);

				break;
			}

			INSIST(iSeg != this->size() - 2); //should have found the right segment by now
		}
	}

	return y;
}


/*
 * Catmul rom spline computation: http://www.mvps.org/directx/articles/catmull/
 */
double Curve::getCurveResponsePrecise(const CurvePoint &p0, const CurvePoint &p1, 
			 			              const CurvePoint &p2, const CurvePoint &p3,
						              double x, double xDistThres) const
{
	INSIST(p0.x <= p1.x);
	INSIST(p1.x <= p2.x);
	INSIST(p2.x <= p3.x);
	CHECK_RANGE(x, 0.0, 1.0);

	double termX0 = 0.5 * (2.0 * p1.x);
	double termX1 = 0.5 * (p2.x - p0.x);
	double termX2 = 0.5 * ((2.0 * p0.x) - (5.0 * p1.x) + (4.0 * p2.x) - p3.x);
	double termX3 = 0.5 * ((3.0 * p1.x) - (3.0 * p2.x) + p3.x - p0.x);

	double termY0 = 0.5 * (2.0 * p1.y);
	double termY1 = 0.5 * (p2.y - p0.y);
	double termY2 = 0.5 * ((2.0 * p0.y) - (5.0 * p1.y) + (4.0 * p2.y) - p3.y);
	double termY3 = 0.5 * ((3.0 * p1.y) - (3.0 * p2.y) + p3.y - p0.y);

	double t_Left   = 0.0;
	double t_Right  = 2.0;
	double t_Middle = 1.0;

	double t_LeftX   = 0.0, t_LeftY;
	double t_RightX  = 2.0, t_RightY;
	double t_MiddleX = 1.0, t_MiddleY;

	double y = DBL_MAX;

	while(y == DBL_MAX)
	{
		if(InRange(x, t_LeftX, t_MiddleX))
		{
			t_Right  = t_Middle;
			t_Middle = (t_Left + t_Right) / 2.0;
		}
		else
		{
			CHECK_RANGE(x, t_MiddleX, t_RightX);
			t_Left   = t_Middle;
			t_Middle = (t_Left + t_Right) / 2.0;
		}

		for(int i = 0; i < 3; i++)
		{
			double t;
			double *t_X, *t_Y;

			switch(i)
			{
			case 0:
				t   = t_Left;
				t_X = &t_LeftX;
				t_Y = &t_LeftY;
				break;
			case 1:
				t   = t_Middle;
				t_X = &t_MiddleX;
				t_Y = &t_MiddleY;
				break;
			case 2:
				t   = t_Right;
				t_X = &t_RightX;
				t_Y = &t_RightY;
				break;
			default:
				REPORT_FAILURE("Unknown case");
				break;
			}

			double t2 = t * t;
			double t3 = t2 * t;

			*t_X = termX0 + (termX1 * t) + (termX2 * t2) + (termX3 * t3);
			*t_Y = termY0 + (termY1 * t) + (termY2 * t2) + (termY3 * t3);

			*t_X = max( min(*t_X, 1.0), 0.0 );
			*t_Y = max( min(*t_Y, 1.0), 0.0 );

			if(fabs(*t_X - x) < xDistThres)
			{
				y = *t_Y;
				break;
			}
		}
	} 

	y = (y * (this->maxOut - this->minOut)) + this->minOut;
	return y;
}

